import functools
import inspect
import operator

import numpy as np
import scipy.constants

import pyoperators
from pyoperators import (
    BlockDiagonalOperator,
    Cartesian2SphericalOperator,
    CompositionOperator,
    DenseBlockDiagonalOperator,
    DenseOperator,
    DiagonalNumexprOperator,
    DiagonalOperator,
    HomothetyOperator,
    IdentityOperator,
    Operator,
    Spherical2CartesianOperator,
)
from pyoperators.flags import inplace, linear, orthogonal, real, separable, square
from pyoperators.memory import empty, ones
from pyoperators.utils import (
    float_dtype,
    isscalarlike,
    operation_assignment,
    product,
    strenum,
    tointtuple,
)

from . import _flib as flib
from .datatypes import FitsArray, Map
from .quantities import Quantity, _divide_unit, _multiply_unit
from .sparse import (
    FSRBlockMatrix,
    FSRMatrix,
    FSRRotation2dMatrix,
    FSRRotation3dMatrix,
    SparseOperator,
)
from .wcsutils import create_fitsheader

__all__ = [
    'BlackBodyOperator',
    'CartesianEquatorial2GalacticOperator',
    'CartesianGalactic2EquatorialOperator',
    'CartesianEquatorial2HorizontalOperator',
    'CartesianHorizontal2EquatorialOperator',
    'ConvolutionTruncatedExponentialOperator',
    'PointingMatrix',
    'PowerLawOperator',
    'ProjectionOperator',
    'ProjectionInMemoryOperator',
    'ProjectionOnFlyOperator',
    'RollOperator',
    'SphericalEquatorial2GalacticOperator',
    'SphericalGalactic2EquatorialOperator',
    'SphericalEquatorial2HorizontalOperator',
    'SphericalHorizontal2EquatorialOperator',
]


def block_diagonal(*partition_args, **keywords):
    """
    Class decorator that partitions an Operator along a specified axis.
    It adds a 'partitionin' keyed argument to the class constructor.

    Subclasses can also benefit from a decorated class, as long as they do not
    define an __init__ method.

    Parameters
    ----------
    axisin : int
        Input partition axis for chunk partitioning.
    new_axisin : int
        Input partition axis for stack partitioning.

    *partition_args: string varargs
       Specify the class' arguments of the __init__ method that  can be a
       sequence to be dispatched to the partitioned operators.

    Example
    -------
    >>> @block_diagonal('value', axisin=0)
    >>> @linear
    >>> class MyOp(Operator):
    >>>     def __init__(self, value, shapein=None):
    >>>         self.value = value
    >>>         Operator.__init__(self, lambda i,o: np.multiply(i,value,o),
    >>>                           shapein=shapein)
    >>>
    >>> op = MyOp([1, 2, 3], shapein=3, partitionin=(1, 1, 1))
    >>> op.todense()
    array([[ 1.,  0.,  0.],
           [ 0.,  2.,  0.],
           [ 0.,  0.,  3.]])

    >>> [o.value for o in op.operands]
    [1, 2, 3]

    """
    # the following way to extract keywords is unnecessary in Python3 (PEP3102)
    if 'axisin' not in keywords and 'new_axisin' not in keywords:
        raise TypeError("Missing 'axisin' or 'new_axisin' keyword.")

    if 'axisin' in keywords:
        axisin = keywords['axisin']
        new_axisin = None
        del keywords['axisin']
    else:
        axisin = None
        new_axisin = keywords['new_axisin']
        del keywords['new_axisin']

    if len(keywords) > 0:
        raise TypeError('Invalid keyed argument.')

    def func(cls):
        @functools.wraps(cls.__init__)
        def partition_init(self, *args, **keywords):

            # get number of blocks through the partitionin keyword
            n1 = 0
            partitionin = None
            if 'partitionin' in keywords:
                if keywords['partitionin'] is not None:
                    partitionin = tointtuple(keywords['partitionin'])
                    n1 = len(partitionin)
                del keywords['partitionin']

            # get __init__ argument names, except self
            class_args = inspect.getfullargspec(cls.__init_original__).args
            class_args.pop(0)

            # get number of blocks through the arguments
            ns = [
                0 if isscalarlike(a) else len(a)
                for i, a in enumerate(args)
                if class_args[i] in partition_args
            ] + [
                0 if isscalarlike(v) else len(v)
                for k, v in keywords.items()
                if k in partition_args
            ]

            n2 = 0 if len(ns) == 0 else max(ns)
            if any(n not in (0, n2) for n in ns):
                raise ValueError(
                    'The partition variables do not have the same'
                    ' number of elements.'
                )

            # bail if no partitioning is found
            n = max(n1, n2)
            if n == 0:
                cls.__init_original__(self, *args, **keywords)
                return

            # check the two methods are compatible
            if n1 != 0 and n2 != 0 and n1 != n2:
                raise ValueError(
                    'The specified partitioning is incompatible w'
                    'ith the number of elements in the partition '
                    'variables.'
                )

            # Implicit partition
            if partitionin is None:
                if axisin is not None:
                    partitionin = n * (None,)
                else:
                    partitionin = n * (1,)

            # dispatch arguments
            n = len(partitionin)
            argss = tuple(
                tuple(
                    (
                        a[i]
                        if class_args[j] in partition_args and not isscalarlike(a)
                        else a
                    )
                    for j, a in enumerate(args)
                )
                for i in range(n)
            )
            keyss = tuple(
                dict(
                    (k, v[i]) if k in partition_args and not isscalarlike(v) else (k, v)
                    for k, v in keywords.items()
                )
                for i in range(n)
            )

            # the input shapein/out describe the BlockDiagonalOperator
            def reshape(s, p, a, na):
                s = list(tointtuple(s))
                if a is not None:
                    s[a] = p
                else:
                    np.delete(s, na)
                return tuple(s)

            if 'shapein' in keywords:
                shapein = keywords['shapein']
                for keys, p in zip(keyss, partitionin):
                    keys['shapein'] = reshape(shapein, p, axisin, new_axisin)
            if 'shapeout' in keywords:
                shapeout = keywords['shapeout']
                for keys, p in zip(keyss, partitionin):
                    keys['shapeout'] = reshape(shapeout, p, axisin, new_axisin)

            # instantiate the partitioned operators
            ops = [cls.__new__(type(self), *a, **k) for a, k in zip(argss, keyss)]
            for o, a, k in zip(ops, argss, keyss):
                if isinstance(o, cls):
                    cls.__init_original__(o, *a, **k)

            self.__class__ = BlockDiagonalOperator
            self.__init__(
                ops, partitionin=partitionin, axisin=axisin, new_axisin=new_axisin
            )

        cls.__init_original__ = cls.__init__
        cls.__init__ = partition_init
        return cls

    return func


@real
class BlackBodyOperator(DiagonalOperator):
    """
    Diagonal operator whose normalised diagonal values are given by the Planck
    equation, optionally modified by a power-law emissivity of given slope.

    In other words, assuming that the set of spectra fnu[i](nu) is proportional
    to the black body law of temperature T[i], and that the flux density of the
    spectra at the reference frequency nu0 is given by fnu_nu0[i], then the
    flux densities fnu_nu[i] at a frequency nu can be determined by using the
    BlackBodyOperator:
        bb = BlackBodyOperator(T, frequency=nu, frequency0=nu0)
        fnu_nu = bb(fnu_nu0)

    Example
    -------
    >>> from pysimulators import gaussian
    >>> T = gaussian((128, 128), sigma=32)  # Temperature map
    >>> ws = np.arange(90, 111) * 1e-6
    >>> bb = [BlackBodyOperator(T, wavelength=w, wavelength0=100e-6)
              for w in ws]

    """

    def __init__(
        self,
        temperature,
        frequency=None,
        frequency0=None,
        wavelength=None,
        wavelength0=None,
        beta=0,
        **keywords,
    ):
        """
        Parameters
        ----------
        temperature : array-like or scalar of float
            Black body temperature, in Kelvin.
        frequency : scalar float
            Frequency, in Hertz, at which black body values will be computed.
        frequency0 : scalar float
            Reference frequency, in Hertz. If the frequency keyword is equal
            to frequency0, the operator is the identity operator.
        wavelength : scalar float
            Wavelength, in meters, at which black body values will be computed.
            (alternative to the frequency keyword.)
        wavelength0 : scalar float
            Reference wavelength, in meters. If the wavelength keyword is equal
            to wavelength0, the operator is the identity operator.
            (alternative to the frequency0 keyword.)
        beta : array-like or scalar of float
            Slope of the emissivity (the spectrum is multiplied by nu**beta)
            The default value is 0 (non-modified black body).

        """
        temperature = np.asarray(temperature, float)
        if frequency is None and wavelength is None:
            raise ValueError(
                'The operating frequency or wavelength is not spe' 'cified.'
            )
        if frequency0 is None and wavelength0 is None:
            raise ValueError(
                'The reference frequency or wavelength is not spe' 'cified.'
            )
        if frequency is not None and wavelength is not None:
            raise ValueError('Ambiguous operating frequency / wavelength.')
        if frequency0 is not None and wavelength0 is not None:
            raise ValueError('Ambiguous reference frequency / wavelength.')

        c = scipy.constants.c
        h = scipy.constants.h
        k = scipy.constants.k

        if frequency is not None:
            nu = np.asarray(frequency, float)
        else:
            nu = c / np.asarray(wavelength, float)
        if nu.ndim != 0:
            raise TypeError('The operating frequency or wavelength is not a sc' 'alar.')
        if frequency0 is not None:
            nu0 = np.asarray(frequency0, float)
        else:
            nu0 = c / np.asarray(wavelength0, float)
        if nu.ndim != 0:
            raise TypeError('The operating frequency or wavelength is not a sc' 'alar.')
        beta = float(beta)
        data = (
            (nu / nu0) ** (3 + beta)
            * np.expm1(h * nu0 / (k * temperature))
            / np.expm1(h * nu / (k * temperature))
        )
        DiagonalOperator.__init__(self, data, **keywords)
        self.temperature = temperature
        self.beta = beta
        self.frequency = frequency
        self.frequency0 = frequency0
        self.wavelength = wavelength
        self.wavelength0 = wavelength0


@real
@linear
@square
@inplace
@separable
class ConvolutionTruncatedExponentialOperator(Operator):
    """
    Apply a truncated exponential response to the signal

    Parameters
    ==========
    tau: float or array-like
        Time constant divided by the signal sampling period.

    """

    def __init__(self, tau, dtype=None, **keywords):
        if hasattr(tau, 'SI'):
            tau = tau.SI
            if tau.unit != '':
                raise ValueError('The time constant must be dimensionless.')
        if dtype is None:
            tau = np.asarray(tau)
            if tau.dtype.kind == 'f':
                dtype = tau.dtype
            else:
                dtype = float
        tau = np.asarray(tau, dtype)
        if np.all(tau <= 0):
            self.__class__ = IdentityOperator
            self.__init__(dtype=dtype, **keywords)
            return
        self.tau = tau
        Operator.__init__(self, dtype=dtype, **keywords)

    def direct(self, input, output):
        input_, nd, nt, int_all = _ravel_strided(input)
        output_, nd, nt, ont_all = _ravel_strided(output)
        f = f'trexp_direct_r{input.dtype.itemsize}'
        try:
            getattr(flib.operators, f)(
                input_,
                nd,
                nt,
                int_all,
                output_,
                nt,
                ont_all,
                self.tau.astype(input.dtype),
            )
        except AttributeError:
            raise NotImplementedError()

    def transpose(self, input, output):
        input_, nd, nt, int_all = _ravel_strided(input)
        output_, nd, nt, ont_all = _ravel_strided(output)
        f = f'trexp_transpose_r{input.dtype.itemsize}'
        try:
            getattr(flib.operators, f)(
                input_,
                nd,
                nt,
                int_all,
                output_,
                nt,
                ont_all,
                self.tau.astype(input.dtype),
            )
        except AttributeError:
            raise NotImplementedError()

    def validatein(self, shape):
        if self.tau.size == 1:
            return
        if self.tau.size != product(shape[:-1]):
            raise ValueError(
                f"The shape of the input '{shape}' is incompatible with the number "
                f"of scales '{self.tau.size}'."
            )


def _ravel_strided(x):
    if x.ndim == 1:
        return x, 1, x.size, x.size
    if x.ndim != 2:
        raise NotImplementedError()
    if x.strides[1] != 8:
        raise ValueError('The array is not C-contiguous.')
    shape_all = x.shape[0], x.strides[0] // 8
    x_ = np.lib.stride_tricks.as_strided(x, (product(shape_all),), (8,))
    return x_, x.shape[0], x.shape[1], shape_all[1]


@real
class PowerLawOperator(DiagonalNumexprOperator):
    """
    Diagonal operator which extrapolates an input following the power law:
        output = (x/x0)**alpha * input

    This operator is lightweight, in the sense that it does not copy the
    potentially large argument 'alpha' and that the diagonal elements are
    computed on the fly.

    Parameters
    ----------
    alpha : array-like
        The broadcastable power-law slope.
    x : float
        Abscissa at which the extrapolation is performed.
    x0 : float
        The reference abscissa.

    Example
    -------
    >>> import scipy.constants
    >>> c = scipy.constants.c
    >>> nu0 = c/10e-6
    >>> op = PowerLawOperator(-1, c/11e-6, nu0)
    >>> op([1, 10, 100])
    array([2.75e-11, 2.75e-10, 2.75e-09])
    """

    def __init__(self, alpha, x, x0, scalar=1, **keywords):
        alpha = np.asarray(alpha, float)
        if alpha.ndim > 0:
            keywords['shapein'] = alpha.shape
        x = np.asarray(x, float)
        if x.ndim != 0:
            raise TypeError('The input is not a scalar.')
        x0 = np.asarray(x0, float)
        if x0.ndim != 0:
            raise TypeError('The reference input is not a scalar.')
        scalar = np.asarray(scalar, float)
        if scalar.ndim != 0:
            raise TypeError('The scalar coefficient is not a scalar.')
        if 'dtype' not in keywords:
            keywords['dtype'] = float
        global_dict = {'x': x, 'x0': x0, 's': scalar}
        super().__init__(
            alpha, 's * (x / x0) ** alpha', global_dict, var='alpha', **keywords
        )
        self.alpha = alpha
        self.x = x
        self.x0 = x0
        self.scalar = scalar
        self.set_rule(
            (HomothetyOperator, '.'),
            lambda o, s: PowerLawOperator(s.alpha, s.x, s.x0, o.data * s.scalar),
            CompositionOperator,
        )

    @staticmethod
    def _rule_block(self, op, shape, partition, axis, new_axis, func_operation):
        return DiagonalOperator._rule_block(
            self,
            op,
            shape,
            partition,
            axis,
            new_axis,
            func_operation,
            self.x,
            self.x0,
            scalar=self.scalar,
        )

    def __str__(self):
        return f'powerlaw(..., α={self.alpha}, x/x0={self.x / self.x0})'


class PointingMatrix(FitsArray):
    default_dtype = np.dtype([('index', np.int32), ('value', np.float32)])

    def __new__(cls, array, shape_input, copy=True, ndmin=0):
        result = FitsArray.__new__(cls, array, copy=copy, ndmin=ndmin).view(
            cls.default_dtype, cls
        )
        result.header = create_fitsheader(result.shape[::-1])
        result.shape_input = shape_input
        return result

    def __array_finalize__(self, array):
        FitsArray.__array_finalize__(self, array)
        self.shape_input = getattr(array, 'shape_input', None)

    @classmethod
    def empty(cls, shape, shape_input, verbose=True):
        buffer = empty(
            shape,
            cls.default_dtype,
            verbose=verbose,
            description='for the pointing matrix',
        )
        return PointingMatrix(buffer, shape_input, copy=False)

    @classmethod
    def ones(cls, shape, shape_input, verbose=True):
        buffer = empty(
            shape,
            cls.default_dtype,
            verbose=verbose,
            description='for the pointing matrix',
        )
        buffer['index'] = -1
        buffer['value'] = 1
        return PointingMatrix(buffer, shape_input, copy=False)

    @classmethod
    def zeros(cls, shape, shape_input, verbose=True):
        buffer = empty(
            shape,
            cls.default_dtype,
            verbose=verbose,
            description='for the pointing matrix',
        )
        buffer['index'] = -1
        buffer['value'] = 0
        return PointingMatrix(buffer, shape_input, copy=False)

    def isvalid(self):
        """
        Return true if the indices in the pointing matrix are greater or equal
        to -1 and lesser than the number of pixels in the map.

        """
        if self.size == 0:
            return True
        npixels = product(self.shape_input)
        result = flib.pointingmatrix.isvalid(
            self.ravel().view(np.int64),
            self.shape[-1],
            product(self.shape[:-1]),
            npixels,
        )
        return bool(result)

    def pack(self, mask):
        """
        Inplace renumbering of the pixel indices in the pointing matrix,
        discarding masked or unobserved pixels.

        """
        if self.size == 0:
            return
        flib.pointingmatrix.pack(
            self.ravel().view(np.int64),
            mask.view(np.int8).T,
            self.shape[-1],
            self.size // self.shape[-1],
        )
        self.shape_input = tointtuple(mask.size - np.sum(mask))

    def get_mask(self, out=None):
        if out is None:
            out = empty(self.shape_input, np.bool_, description='as new mask')
            out[...] = True
        elif out.dtype != bool:
            raise TypeError('The output mask argument has an invalid type.')
        elif out.shape != self.shape_input:
            raise ValueError('The output mask argument has an incompatible sha' 'pe.')
        if self.size == 0:
            return out
        flib.pointingmatrix.mask(
            self.ravel().view(np.int64),
            out.view(np.int8).T,
            self.shape[-1],
            self.size // self.shape[-1],
        )
        return out


@real
@linear
class ProjectionBaseOperator(Operator):
    """
    Abstract class for projection operators.

    """

    def __init__(
        self, units=None, derived_units=None, attrin={}, attrout={}, **keywords
    ):

        if units is None:
            units = ('', '')
        if derived_units is None:
            derived_units = ({}, {})

        unit = _divide_unit(Quantity(1, units[0])._unit, Quantity(1, units[1])._unit)

        Operator.__init__(
            self,
            dtype=float,
            attrin=self.set_attrin,
            attrout=self.set_attrout,
            **keywords,
        )
        self._attrin = attrin
        self._attrout = attrout
        self.unit = unit
        self.duout, self.duin = derived_units

    def direct(self, input, output):
        matrix = self.matrix
        if matrix.size == 0:
            return
        if not output.flags.contiguous:
            if pyoperators.memory.verbose:
                print('Optimize me: Projection output is not contiguous.')
            output_ = np.empty_like(output)
        else:
            output_ = output
        flib.pointingmatrix.direct(
            matrix.ravel().view(np.int64), input.T, output_.T, matrix.shape[-1]
        )
        if not output.flags.contiguous:
            output[...] = output_

    def transpose(self, input, output, operation=operation_assignment):
        matrix = self.matrix
        if operation is operation_assignment:
            output[...] = 0
        elif operation is not operator.iadd:
            raise ValueError('Invalid reduction operation.')
        if matrix.size == 0:
            return
        if not input.flags.contiguous:
            if pyoperators.config.VERBOSE:
                print('Optimize me: Projection.T input is not contiguous.')
            input_ = np.ascontiguousarray(input)
        else:
            input_ = input
        flib.pointingmatrix.transpose(
            matrix.ravel().view(np.int64), input_.T, output.T, matrix.shape[-1]
        )

    def set_attrin(self, attr):
        if '_header' in attr:
            del attr['_header']
        unitout = attr['_unit'] if '_unit' in attr else {}
        if unitout:
            attr['_unit'] = _divide_unit(unitout, self.unit)
        if self.duin is not None:
            attr['_derived_units'] = self.duin
        attr.update(self._attrin)

    def set_attrout(self, attr):
        if '_header' in attr:
            del attr['_header']
        unitin = attr['_unit'] if '_unit' in attr else {}
        if unitin:
            attr['_unit'] = _multiply_unit(unitin, self.unit)
        if self.duout is not None:
            attr['_derived_units'] = self.duout
        attr.update(self._attrout)

    def apply_mask(self, mask):
        """
        Set pointing matrix values to zero following an input mask of shape
        the operator output shape.

        """
        raise NotImplementedError()

    def get_mask(self, out=None):
        return self.matrix.get_mask(out=out)

    def get_pTp(self, out=None):
        matrix = self.matrix
        npixels = product(matrix.shape_input)
        if out is None:
            out = empty((npixels, npixels), bool, description='for pTp array')
            out[...] = 0
        elif out.dtype != np.bool_:
            raise TypeError('The output ptp argument has an invalid type.')
        elif out.shape != (npixels, npixels):
            raise ValueError('The output ptp argument has an incompatible shap' 'e.')
        if matrix.size == 0:
            return out
        flib.pointingmatrix.ptp(
            matrix.ravel().view(np.int64),
            out.T,
            matrix.shape[-1],
            matrix.size // matrix.shape[-1],
            npixels,
        )
        return out

    def get_pTx_pT1(self, x, out=None, mask=None):
        """
        Return a tuple of two arrays: the transpose of the projection
        applied over the input and over one. In other words, it returns
        the back projection of the input and its coverage.

        """
        matrix = self.matrix
        if out is None:
            shape = matrix.shape_input
            out = Map.zeros(shape), Map.zeros(shape)
            self.set_attrin(out[0].__dict__)
            if hasattr(x, 'unit'):
                out[0].unit = x.unit
        if matrix.size == 0:
            return out
        mask = mask or getattr(x, 'mask', None)
        if mask is not None:
            flib.pointingmatrix.backprojection_weight_mask(
                matrix.ravel().view(np.int64),
                x.ravel(),
                x.mask.ravel().view(np.int8),
                out[0].ravel(),
                out[1].ravel(),
                matrix.shape[-1],
            )
        else:
            flib.pointingmatrix.backprojection_weight(
                matrix.ravel().view(np.int64),
                x.ravel(),
                out[0].ravel(),
                out[1].ravel(),
                matrix.shape[-1],
            )
        return out

    def intersects(self, fitscoords, axis=None, f=None):
        """
        Return True if a map pixel is seen by the pointing matrix.

        """
        matrix = self.matrix
        axes = (1,) + matrix.shape_input[:0:-1]
        index = np.array(fitscoords) - 1
        index = int(np.sum(index * np.cumproduct(axes)))

        shape = matrix.shape
        matrix_ = matrix.ravel().view(np.int64)
        if f is not None:
            out = f(matrix_, index, shape[-1], shape[-2], shape[-3])
        elif axis is None:
            out = flib.pointingmatrix.intersects(
                matrix_, index, shape[-1], shape[-2], shape[-3]
            )
        elif axis == 0 or axis == -3:
            out = flib.pointingmatrix.intersects_axis3(
                matrix_, index, shape[-1], shape[-2], shape[-3]
            )
        elif axis == 1 or axis == -2:
            out = flib.pointingmatrix.intersects_axis2(
                matrix_, index, shape[-1], shape[-2], shape[-3]
            )
        else:
            raise ValueError('Invalid axis.')

        if isinstance(out, np.ndarray):
            out = out.view(bool)
        else:
            out = bool(out)
        return out


class ProjectionInMemoryOperator(ProjectionBaseOperator):
    """
    Projection operator that stores the pointing matrix in memory.

    Attributes
    ----------
    matrix : composite dtype ('value', 'f4'), ('index', 'i4')
        The pointing matrix.

    """

    def __init__(self, matrix, units=None, derived_units=None, **keywords):
        """
        Parameters
        ----------
        matrix : PointingMatrix
            The pointing matrix.
        units : unit
            Unit of the pointing matrix. It is used to infer the output unit
            from that of the input.
        derived_units : tuple (derived_units_out, derived_units_in)
            Derived units to be added to the direct or transposed outputs.

        """
        if 'shapein' in keywords:
            raise TypeError("The 'shapein' keyword should not be used.")
        shapein = matrix.shape_input

        if 'shapeout' in keywords:
            raise TypeError("The 'shapeout' keyword should not be used.")
        shapeout = matrix.shape[:-1]

        ProjectionBaseOperator.__init__(
            self,
            shapein=shapein,
            shapeout=shapeout,
            units=units,
            derived_units=derived_units,
            **keywords,
        )
        self.matrix = matrix
        self.set_rule('T,.', self._rule_ptp, CompositionOperator)
        self.set_rule((DiagonalOperator, '.'), self._rule_diagonal, CompositionOperator)

    def apply_mask(self, mask):
        mask = np.asarray(mask, np.bool_)
        matrix = self.matrix
        if mask.shape != self.shapeout:
            raise ValueError(
                f"The mask shape '{mask.shape}' is incompatible with that o"
                f"f the pointing matrix '{matrix.shape}'."
            )
        matrix.value.T[...] *= 1 - mask.T

    @staticmethod
    def _rule_ptp(pt, p):
        if p.matrix.shape[-1] != 1:
            return
        cov = np.zeros(p.shapein)
        flib.pointingmatrix.ptp_one(
            p.matrix.ravel().view(np.int64), cov.ravel(), p.matrix.size, cov.size
        )
        return DiagonalOperator(cov)

    @staticmethod
    def _rule_diagonal(d, self):
        # rightward broadcasting can fail with structured arrays (np 1.7)
        if self.matrix.shape[-1] != 1 or d.broadcast == 'rightward':
            return
        matrix = self.matrix.copy()
        result = ProjectionInMemoryOperator(matrix)
        matrix.value *= d.get_data()
        return result


class ProjectionOnFlyOperator(ProjectionBaseOperator):
    """
    Projection operator that recomputes the pointing matrix on the fly.

    ProjectionOnFlyOperator instances are grouped (usually organised as a block
    column operator) and they share a dictionary, which holds one pointing
    matrix and a hashable to identify with which element of the group this
    pointing matrix is associated. If another element requests the
    computation of its pointing matrix, the common dict is updated accordingly,
    and the pointing matrix is substituted.

    """

    def __init__(
        self, place_holder, id, func, units=None, derived_units=None, **keywords
    ):
        """
        Parameters
        ----------
        place_holder : dict
            dict whose keys 'matrix' contains the pointing matrix and 'id' the
            hashable to identify with which operator it is associated.
        id : hashable
            The on-fly projection operator identifier.
        func : callable
            Function used to compute the pointing matrix on the fly:
                 matrix = func(id)
        units : unit
            Unit of the pointing matrix. It is used to infer the output unit
            from that of the input.
        derived_units : tuple (derived_units_out, derived_units_in)
            Derived units to be added to the direct or transposed outputs.

        """
        if 'matrix' not in place_holder:
            place_holder['matrix'] = None
            place_holder['id'] = None
        self.place_holder = place_holder
        self.id = id
        self.func = func
        ProjectionBaseOperator.__init__(
            self, units=units, derived_units=derived_units, **keywords
        )

    @property
    def matrix(self):
        if self.place_holder['id'] is self.id:
            return self.place_holder['matrix']
        self.place_holder['matrix'] = None
        matrix = self.func(self.id)
        self.place_holder['id'], self.place_holder['matrix'] = self.id, matrix
        return matrix


class ProjectionOperator(SparseOperator):
    """
    Projection operator. It is a SparseOperator with convenience methods.

    Attributes
    ----------
    matrix : FSRMatrix, FSRBlockMatrix, FSRRotation2dMatrix or FSRRotation3dMatrix
        The projection sparse matrix.
    """

    def __init__(self, arg, **keywords):
        if type(arg) not in (
            FSRMatrix,
            FSRBlockMatrix,
            FSRRotation2dMatrix,
            FSRRotation3dMatrix,
        ):
            raise TypeError('The input sparse matrix type is invalid.')
        self._flib_id = {
            FSRMatrix: '',
            FSRBlockMatrix: '_block',
            FSRRotation2dMatrix: '_rot2d',
            FSRRotation3dMatrix: '_rot3d',
        }[type(arg)]
        SparseOperator.__init__(self, arg, **keywords)

    def canonical_basis_in_kernel_old(self, out=None, operation=operation_assignment):
        """
        Return a boolean array whose values are True if the component of
        the canonical basis is in the operator's kernel.

        """
        data = self.matrix.data
        shape = self.shapein
        if shape is None:
            shape = (self.matrix.shape[1],)
        elif self.matrix.block_shape[1] > 1:
            shape = shape[:-1]
        if out is None:
            if operation is not operation_assignment:
                raise ValueError('The array for inplace operation is not specified')
            out = empty(shape, bool)
        elif out.dtype != bool:
            raise TypeError("The keyword 'out' has an invalid type.")
        elif out.shape != shape:
            raise ValueError("The keyword 'out' has an incompatible shape.")
        if data.size == 0:
            operation(out, True)
            return out

        i, m = (int(_.dtype.itemsize) for _ in (data.index, self.matrix))
        f = f'fsr{self._flib_id}_kernel_i{i}_r{m}'
        n = product(shape)
        if hasattr(flib.operators, f):
            if operation in (operation_assignment, operator.iand, operator.imul):
                if operation is operation_assignment:
                    out[...] = True
                kernel = out
            else:
                kernel = np.ones_like(out)

            func = getattr(flib.operators, f)
            func(
                data.view(np.int8).ravel(),
                kernel.view(np.int8).ravel(),
                self.matrix.ncolmax,
                data.shape[0],
            )
            if operation in (operation_assignment, operator.iand, operator.imul):
                return out
        else:
            if isinstance(self.matrix, FSRMatrix):
                indices = data.index[data.value != 0]
            elif isinstance(self.matrix, FSRRotation2dMatrix):
                indices = data.index[(data.r11 != 0) | (data.r21 != 0)]
            else:
                indices = data.index[data.r11 != 0]
            kernel = np.histogram(indices, n, range=(0, n))[0] == 0
        operation(out, kernel)
        return out

    def canonical_basis_in_kernel(self, out=None, operation=operation_assignment):
        """
        Return a boolean array whose values are True if the component of
        the canonical basis is in the operator's kernel.

        """
        data = self.matrix.data
        shape = self.shapein
        if shape is None:
            shape = (self.matrix.shape[1],)
        elif self.matrix.block_shape[1] > 1:
            shape = shape[:-1]
        if out is None:
            if operation is not operation_assignment:
                raise ValueError('The array for inplace operation is not specified')
            out = empty(shape, bool)
        elif out.dtype != bool:
            raise TypeError("The keyword 'out' has an invalid type.")
        elif out.shape != shape:
            raise ValueError("The keyword 'out' has an incompatible shape.")
        if data.size == 0:
            operation(out, True)
            return out

        itype = data.dtype['index']
        ftype = data.dtype[1]
        if itype.type in (np.int8, np.int16, np.int32, np.int64) and ftype.type in (
            np.float32,
            np.float64,
        ):
            if operation in (operation_assignment, operator.iand, operator.imul):
                if operation is operation_assignment:
                    out[...] = True
                kernel = out
            else:
                kernel = np.ones_like(out)

            f = f'fsr_kernel_i{int(itype.itemsize)}'
            func = getattr(flib.sparse, f)
            func(
                data.view(np.int8).ravel(),
                kernel.view(np.int8).ravel(),
                self.matrix.ncolmax,
                product(data.shape[:-1]),
                data.strides[-1],
            )
            if operation in (operation_assignment, operator.iand, operator.imul):
                return out
        else:
            if isinstance(self.matrix, FSRMatrix):
                indices = data.index[data.value != 0]
            elif isinstance(self.matrix, FSRRotation2dMatrix):
                indices = data.index[(data.r11 != 0) | (data.r21 != 0)]
            else:
                indices = data.index[data.r11 != 0]
            n = product(shape)
            kernel = np.histogram(indices, n, range=(0, n))[0] == 0
        operation(out, kernel)
        return out

    def pT1(self, out=None, operation=operation_assignment):
        """
        Return the transpose of the projection over one. In other words,
        it returns the projection coverage.

        """
        shapeout = self.shapein
        if shapeout is None:
            shapeout = (self.matrix.shape[1],)
        elif self.matrix.block_shape[1] > 1:
            shapeout = shapeout[:-1]
        if out is None:
            out = empty(shapeout, self.dtype)
        elif not isinstance(out, np.ndarray):
            raise TypeError('The output keyword is not an ndarray.')
        elif out.shape != shapeout:
            raise ValueError(
                f'The output arrays have a shape incompatible with'
                f" the expected one '{shapeout}'."
            )
        if operation not in (operation_assignment, operator.iadd):
            raise ValueError('Invalid reduction operation.')

        isize = int(self.matrix.data.index.dtype.itemsize)
        rsize = int(self.matrix.dtype.itemsize)
        vsize = int(out.dtype.itemsize)
        f = f'fsr{self._flib_id}_pt1_i{isize}_r{rsize}_v{vsize}'
        if hasattr(flib.operators, f):
            if operation is operation_assignment:
                out[...] = 0
            func = getattr(flib.operators, f)
            func(
                self.matrix.data.view(np.int8).ravel(),
                out.ravel(),
                self.matrix.ncolmax,
                self.matrix.data.shape[0],
            )
        else:
            if self.shapeout is None:
                shapein = (self.matrix.shape[0],)
            else:
                shapein = self.shapeout
            if isinstance(self.matrix, FSRMatrix):
                x = ones(shapein, self.dtype)
                self.T(x, out=out, operation=operation)
            elif isinstance(self.matrix, FSRRotation2dMatrix):
                if operation is operation_assignment:
                    out[...] = 0
                index = self.matrix.data.index.ravel()
                data = np.sqrt(
                    self.matrix.data.r11**2 + self.matrix.data.r21**2
                ).ravel()
                for i, d in zip(index, data):
                    if i >= 0:
                        out[i] += d
            else:
                x = ones(shapein, self.dtype)
                operation(out, self.T(x)[..., 0])
        return out

    def pTx_pT1(self, x, out=None, operation=operation_assignment):
        """
        Return a tuple of two arrays: the transpose of the projection
        applied over the input and over one. In other words, it returns
        the back projection of the input and its coverage.

        """
        if isinstance(self.matrix, FSRRotation2dMatrix):
            raise NotImplementedError(
                'This method does not make sense for FSRRotation2dMatrix spars'
                'e storage.'
            )
        x = np.asarray(x, order='C')
        shapein = self.shapeout
        if shapein is None:
            block_size = self.matrix.block_shape[1]
            shapein = (self.matrix.shape[0],) + (
                () if block_size == 1 else (block_size,)
            )
        shapeout = self.shapein
        if shapeout is None:
            shapeout = (self.matrix.shape[1],)
        elif self.matrix.block_shape[1] > 1:
            shapeout = shapeout[:-1]
        if x.shape != shapein:
            raise ValueError(
                f"Invalid input shape '{x.shape}'. Expected shape is '{shapein}'."
            )
        dtypeout = max(self.dtype, x.dtype)
        if out is None:
            pTx = empty(shapeout, dtypeout)
            pT1 = empty(shapeout, dtypeout)
        elif not isinstance(out, (list, tuple)) or len(out) != 2:
            raise TypeError('The out keyword is not a 2-tuple.')
        elif out[0].dtype != dtypeout or out[1].dtype != dtypeout:
            raise TypeError(
                f'The output arrays have a dtype incompatible with the expected one '
                f"'{dtypeout.type}'."
            )
        elif out[0].shape != shapeout or out[1].shape != shapeout:
            raise ValueError(
                f'The output arrays have a shape incompatible with the expected one '
                f"'{shapeout}'."
            )
        else:
            pTx, pT1 = out
        if operation not in (operation_assignment, operator.iadd):
            raise ValueError('Invalid reduction operation.')

        isize = int(self.matrix.data.index.dtype.itemsize)
        rsize = int(self.matrix.dtype.itemsize)
        vsize = int(dtypeout.itemsize)
        f = f'fsr{self._flib_id}_ptx_pt1_i{isize}_r{rsize}_v{vsize}'
        if hasattr(flib.operators, f):
            if operation is operation_assignment:
                pTx[...] = 0
                pT1[...] = 0
            x = np.asarray(x, dtype=dtypeout, order='C')
            func = getattr(flib.operators, f)
            func(
                self.matrix.data.view(np.int8).ravel(),
                x.T,
                pTx.ravel(),
                pT1.ravel(),
                self.matrix.ncolmax,
            )
        else:
            self.pT1(out=pT1, operation=operation)
            if isinstance(self.matrix, FSRMatrix):
                self.T(x, out=pTx, operation=operation)
            else:
                operation(pTx, self.T(x)[..., 0])
        return pTx, pT1


@orthogonal
@inplace
class RollOperator(Operator):
    def __init__(self, n, axis=None, **keywords):
        Operator.__init__(self, **keywords)
        if axis is None:
            axis = -1 if isscalarlike(n) else tuple(range(-len(n), 0))
        self.axis = (axis,) if isscalarlike(axis) else tuple(axis)
        self.n = (n,) * len(self.axis) if isscalarlike(n) else tuple(n)
        if len(self.axis) != len(self.n):
            raise ValueError(
                'There is a mismatch between the number of axes a' 'nd offsets.'
            )

    def direct(self, input, output):
        output[...] = np.roll(input, self.n[0], axis=self.axis[0])
        for n, axis in zip(self.n[1:], self.axis[1:]):
            output[...] = np.roll(output, n, axis=axis)

    def transpose(self, input, output):
        output[...] = np.roll(input, -self.n[0], axis=self.axis[0])
        for n, axis in zip(self.n[1:], self.axis[1:]):
            output[...] = np.roll(output, -n, axis=axis)


@orthogonal
class _CartesianEquatorialGalactic(DenseOperator):
    _g2e = np.linalg.qr(
        [
            [-0.0548755604, 0.4941094279, -0.8676661490],
            [-0.8734370902, -0.4448296300, -0.1980763734],
            [-0.4838350155, 0.7469822445, 0.4559837762],
        ]
    )[0]


class CartesianEquatorial2GalacticOperator(_CartesianEquatorialGalactic):
    """
    ICRS-to-Galactic cartesian coordinate transform.

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0)
        - the Earth North pole of coordinates (0, 0, 1)

    The galactic direct referential is defined by:
        - the Sun center as the origin
        - the galactic center of coordinates (1, 0, 0)
        - the galactic pole of coordinates (0, 0, 1)

    Note that the equatorial-to-galactic conversion is considered to be
    a rotation, so we neglect the Earth-to-Sun origin translation.

    The galactic pole and center in the ICRS frame are defined by
    (Hipparcos 1997):
        αp = 192.85948°
        δp = 27.12825°
        αc = 266.40510°
        δc = -28.936175°

    The last dimension of the inputs/outputs must be 3.

    Example
    -------
    >>> op = CartesianEquatorial2GalacticOperator()
    >>> op([0, 0, 1])
    array([-0.48383502,  0.74698224,  0.45598378])

    """

    def __init__(self, **keywords):
        _CartesianEquatorialGalactic.__init__(self, self._g2e.T, **keywords)
        self.set_rule('T', lambda s: CartesianGalactic2EquatorialOperator())
        self.set_rule(
            ('.', CartesianGalactic2EquatorialOperator), '1', CompositionOperator
        )


class CartesianGalactic2EquatorialOperator(_CartesianEquatorialGalactic):
    """
    Galactic-to-ICRS cartesian coordinate transform.

    The galactic direct referential is defined by:
        - the Sun center as the origin
        - the galactic center of coordinates (1, 0, 0)
        - the galactic pole of coordinates (0, 0, 1)

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0)
        - the Earth North pole of coordinates (0, 0, 1)

    Note that the galactic-to-equatorial conversion is considered to be
    a rotation, so we neglect the Sun-to-Earth origin translation.

    The galactic pole and center in the ICRS frame are defined by
    (Hipparcos 1997):
        αp = 192.85948°
        δp = 27.12825°
        αc = 266.40510°
        δc = -28.936175°

    The last dimension of the inputs/outputs must be 3.

    Example
    -------
    >>> op = CartesianGalactic2EquatorialOperator()
    >>> op([0, 0, 1])
    array([-0.86766615, -0.19807637,  0.45598378])

    """

    def __init__(self, **keywords):
        _CartesianEquatorialGalactic.__init__(self, self._g2e, **keywords)
        self.set_rule('T', lambda s: CartesianEquatorial2GalacticOperator())
        self.set_rule(
            ('.', CartesianEquatorial2GalacticOperator), '1', CompositionOperator
        )


class _CartesianEquatorialHorizontal(DenseBlockDiagonalOperator):
    def __init__(
        self, convention, time, latitude, longitude, transpose, dtype=None, **keywords
    ):
        conventions = ('NE',)  # 'SW', 'SE')
        if convention not in conventions:
            raise ValueError(
                f'Invalid azimuth angle convention {convention!r}. Expected values are '
                f'{strenum(conventions)}.'
            )
        lst = self._gst2lst(self._jd2gst(time.jd), longitude)
        lst = np.radians(lst * 15)
        latitude = np.radians(latitude)
        if dtype is None:
            dtype = np.result_type(float_dtype(lst.dtype), float_dtype(latitude.dtype))
        slat = np.sin(latitude)
        clat = np.cos(latitude)
        slst = np.sin(lst)
        clst = np.cos(lst)
        m = np.empty(np.broadcast(latitude, lst).shape + (3, 3), dtype)
        m[..., 0, 0] = -slat * clst
        m[..., 0, 1] = -slat * slst
        m[..., 0, 2] = clat
        m[..., 1, 0] = -slst
        m[..., 1, 1] = clst
        m[..., 1, 2] = 0
        m[..., 2, 0] = clat * clst
        m[..., 2, 1] = clat * slst
        m[..., 2, 2] = slat
        if transpose:
            m = m.swapaxes(-1, -2)
        keywords['flags'] = self.validate_flags(
            keywords.get('flags', {}), orthogonal=True
        )
        DenseBlockDiagonalOperator.__init__(self, m, **keywords)

    @staticmethod
    def _jd2gst(jd):
        """
        Convert Julian Dates into Greenwich Sidereal Time.

        From Duffett-Smith, P. : 1988, Practical astronomy with your
        calculator, Cambridge University Press, third edition

        """
        jd0 = np.floor(jd - 0.5) + 0.5
        T = (jd0 - 2451545) / 36525
        T0 = 6.697374558 + 2400.051336 * T + 0.000025862 * T**2
        T0 %= 24
        ut = (jd - jd0) * 24
        T0 += ut * 1.002737909
        T0 %= 24
        return T0

    @staticmethod
    def _gst2lst(gst, longitude):
        """
        Convert Greenwich Sidereal Time into Local Sidereal Time.
        longitude : Geographic longitude EAST in degrees.
        """
        return (gst + longitude / 15) % 24


class CartesianEquatorial2HorizontalOperator(_CartesianEquatorialHorizontal):
    """
    Conversion between equatorial-to-horizontal cartesian coordinates.

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0)
        - the Earth North pole of coordinates (0, 0, 1)

    The horizontal referential is defined by:
        - the observer geographic position as the origin
        - the azimuth reference (North or South) of coordinates (1, 0, 0)
        - whether the azimuth is measured towards the East or West
        - the zenith of coordinates (0, 0, 1)

    Example
    -------
    >>> from astropy.time import Time, TimeDelta
    >>> t0 = Time(['2000-01-01 00:00:00.0'], scale='utc')
    >>> dt = TimeDelta(np.arange(1000)/10, format='sec')
    >>> lat, lon = 48.853291, 2.348751
    >>> op = CartesianEquatorial2HorizontalOperator('NE', t0 + dt, lat, lon)

    """

    def __init__(self, convention, time, latitude, longitude, **keywords):
        """
        convention : 'NE', 'SW', 'SE'
            The azimuth angle convention:
                - 'NE' : measured from the North towards the East (indirect)
                - 'SW' : from the South towards the West (indirect)
                - 'SE' : from the South towards the East (direct)
            But so far, only the 'NE' convention is implemented.
        time : astropy.time.Time
            The observer's time.
        latitude : array-like
            The observer's latitude, in degrees.
        longitude : array-like
            The observation's longitude counted positively eastward,
            in degrees.

        """
        _CartesianEquatorialHorizontal.__init__(
            self, convention, time, latitude, longitude, False, **keywords
        )


class CartesianHorizontal2EquatorialOperator(_CartesianEquatorialHorizontal):
    """
    Conversion between horizontal-to-equatorial cartesian coordinates.

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0)
        - the Earth North pole of coordinates (0, 0, 1)

    The horizontal referential is defined by:
        - the observer geographic position as the origin
        - the azimuth reference (North or South) of coordinates (1, 0, 0)
        - whether the azimuth is measured towards the East or West
        - the zenith of coordinates (0, 0, 1)

    Example
    -------
    >>> from astropy.time import Time, TimeDelta
    >>> t0 = Time(['2000-01-01 00:00:00.0'], scale='utc')
    >>> dt = TimeDelta(np.arange(1000)/10, format='sec')
    >>> lat, lon = 48.853291, 2.348751
    >>> op = CartesianHorizontal2EquatorialOperator('NE', t0 + dt, lat, lon)

    """

    def __init__(self, convention, time, latitude, longitude, **keywords):
        """
        convention : 'NE', 'SW', 'SE'
            The azimuth angle convention:
                - 'NE' : measured from the North towards the East (indirect)
                - 'SW' : from the South towards the West (indirect)
                - 'SE' : from the South towards the East (direct)
            But so far, only the 'NE' convention is implemented.
        time : astropy.time.Time
            The observer's time.
        latitude : array-like
            The observer's latitude, in degrees.
        longitude : array-like
            The observation's longitude counted positively eastward,
            in degrees.

        """
        _CartesianEquatorialHorizontal.__init__(
            self, convention, time, latitude, longitude, True, **keywords
        )


class SphericalEquatorial2GalacticOperator(CompositionOperator):
    """
    ICRS-to-Galactic cartesian coordinate transform.

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0) (zenith=pi/2, az=0)
        - the Earth North pole of coordinates (0, 0, 1) (zenith=0)

    The galactic direct referential is defined by:
        - the Sun center as the origin
        - the galactic center of coordinates (1, 0, 0) (zenith=pi/2, az=0)
        - the galactic pole of coordinates (0, 0, 1) (zenith=0)

    Note that the equatorial-to-galactic conversion is considered to be
    a rotation, so we neglect the Earth-to-Sun origin translation.

    The galactic pole and center in the ICRS frame are defined by
    (Hipparcos 1997):
        αp = 192.85948°
        δp = 27.12825°
        αc = 266.40510°
        δc = -28.936175°

    The last dimension of the inputs/outputs must be 2.

    Example
    -------
    >>> op = SphericalEquatorial2GalacticOperator(degrees=True)
    >>> op([266.40510, -28.936175])
    array([  4.70832664e-05,  -7.91258821e-05])

    """

    def __init__(
        self,
        conventionin='azimuth,elevation',
        conventionout='azimuth,elevation',
        degrees=False,
        **keywords,
    ):
        """
        Parameters
        ----------
        conventionin : str
            One of the following spherical coordinate conventions for
            the input equatorial angles: 'zenith,azimuth', 'azimuth,zenith',
            'elevation,azimuth' and 'azimuth,elevation'.
        conventionout : str
            The spherical coordinate convention for the output galactic angles.
        degrees : boolean, optional
            If true, the angle units are degrees (radians otherwise).

        """
        operands = [
            Cartesian2SphericalOperator(conventionout, degrees=degrees),
            CartesianEquatorial2GalacticOperator(),
            Spherical2CartesianOperator(conventionin, degrees=degrees),
        ]
        CompositionOperator.__init__(self, operands, **keywords)


class SphericalGalactic2EquatorialOperator(CompositionOperator):
    """
    Galactic-to-ICRS cartesian coordinate transform.

    The galactic direct referential is defined by:
        - the Sun center as the origin
        - the galactic center of coordinates (1, 0, 0) (zenith=pi/2, az=0)
        - the galactic pole of coordinates (0, 0, 1) (zenith=0)

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0) (zenith=pi/2, az=0)
        - the Earth North pole of coordinates (0, 0, 1) (zenith=0)

    Note that the galactic-to-equatorial conversion is considered to be
    a rotation, so we neglect the Sun-to-Earth origin translation.

    The galactic pole and center in the ICRS frame are defined by
    (Hipparcos 1997):
        αp = 192.85948°
        δp = 27.12825°
        αc = 266.40510°
        δc = -28.936175°

    The last dimension of the inputs/outputs must be 2.

    Example
    -------
    >>> op = SphericalGalactic2EquatorialOperator(degrees=True)
    >>> op([0, 0])
    array([266.4049948, -28.93617396])

    """

    def __init__(
        self,
        conventionin='azimuth,elevation',
        conventionout='azimuth,elevation',
        degrees=False,
        **keywords,
    ):
        """
        Parameters
        ----------
        conventionin : str
            One of the following spherical coordinate conventions for
            the input galactic angles: 'zenith,azimuth', 'azimuth,zenith',
            'elevation,azimuth' and 'azimuth,elevation'.
        conventionout : str
            The spherical coordinate convention for the output equatorial
            angles.
        degrees : boolean, optional
            If true, the angle units are degrees (radians otherwise).

        """
        operands = [
            Cartesian2SphericalOperator(conventionout, degrees=degrees),
            CartesianGalactic2EquatorialOperator(),
            Spherical2CartesianOperator(conventionin, degrees=degrees),
        ]
        CompositionOperator.__init__(self, operands, **keywords)


class SphericalEquatorial2HorizontalOperator(CompositionOperator):
    """
    Conversion between equatorial-to-horizontal cartesian coordinates.

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0) (zenith=pi/2, az=0)
        - the Earth North pole of coordinates (0, 0, 1) (zenith=0)

    The horizontal referential is defined by:
        - the observer geographic position as the origin
        - the azimuth reference (North or South) (1, 0, 0) (zenith=pi/2, az=0)
        - whether the azimuth is measured towards the East or West
        - the zenith of coordinates (0, 0, 1) (zenith=0)

    The last dimension of the inputs/outputs must be 2.

    Example
    -------
    >>> from astropy.time import Time, TimeDelta
    >>> t0 = Time(['2000-01-01 00:00:00.0'], scale='utc')
    >>> dt = TimeDelta(np.arange(1000)/10, format='sec')
    >>> lat, lon = 48.853291, 2.348751
    >>> op = SphericalEquatorial2HorizontalOperator('NE', t0 + dt, lat, lon)

    """

    def __init__(
        self,
        convention_horizontal,
        time,
        latitude,
        longitude,
        conventionin='azimuth,elevation',
        conventionout='azimuth,elevation',
        degrees=False,
        **keywords,
    ):
        """
        convention_horizontal : 'NE', 'SW', 'SE'
            The azimuth angle convention:
                - 'NE' : measured from the North towards the East (indirect)
                - 'SW' : from the South towards the West (indirect)
                - 'SE' : from the South towards the East (direct)
            But so far, only the 'NE' convention is implemented.
        time : astropy.time.Time
            The observer's time.
        latitude : array-like
            The observer's latitude, in degrees.
        longitude : array-like
            The observation's longitude counted positively eastward,
            in degrees.
        conventionin : str
            One of the following spherical coordinate conventions for
            the input equatorial angles: 'zenith,azimuth', 'azimuth,zenith',
            'elevation,azimuth' and 'azimuth,elevation'.
        conventionout : str
            The spherical coordinate convention for the output horizontal
            angles.
        degrees : boolean, optional
            If true, the angle units are degrees (radians otherwise).

        """
        operands = [
            Cartesian2SphericalOperator(conventionout, degrees=degrees),
            CartesianEquatorial2HorizontalOperator(
                convention_horizontal, time, latitude, longitude
            ),
            Spherical2CartesianOperator(conventionin, degrees=degrees),
        ]
        CompositionOperator.__init__(self, operands, **keywords)


class SphericalHorizontal2EquatorialOperator(CompositionOperator):
    """
    Conversion between horizontal-to-equatorial cartesian coordinates.

    The horizontal referential is defined by:
        - the observer geographic position as the origin
        - the azimuth reference (North or South) (1, 0, 0) (zenith=pi/2, az=0)
        - whether the azimuth is measured towards the East or West
        - the zenith of coordinates (0, 0, 1) (zenith=0)

    The ICRS equatorial direct referential is defined by:
        - the Earth center as the origin
        - the vernal equinox of coordinates (1, 0, 0) (zenith=pi/2, az=0)
        - the Earth North pole of coordinates (0, 0, 1) (zenith=0)

    The last dimension of the inputs/outputs must be 2.

    Example
    -------
    >>> from astropy.time import Time, TimeDelta
    >>> t0 = Time(['2000-01-01 00:00:00.0'], scale='utc')
    >>> dt = TimeDelta(np.arange(1000)/10, format='sec')
    >>> lat, lon = 48.853291, 2.348751
    >>> op = SphericalHorizontal2EquatorialOperator('NE', t0 + dt, lat, lon)

    """

    def __init__(
        self,
        convention_horizontal,
        time,
        latitude,
        longitude,
        conventionin='azimuth,elevation',
        conventionout='azimuth,elevation',
        degrees=False,
        **keywords,
    ):
        """
        convention_horizontal : 'NE', 'SW', 'SE'
            The azimuth angle convention:
                - 'NE' : measured from the North towards the East (indirect)
                - 'SW' : from the South towards the West (indirect)
                - 'SE' : from the South towards the East (direct)
            But so far, only the 'NE' convention is implemented.
        time : astropy.time.Time
            The observer's time.
        latitude : array-like
            The observer's latitude, in degrees.
        longitude : array-like
            The observation's longitude counted positively eastward,
            in degrees.
        conventionin : str
            One of the following spherical coordinate conventions for
            the input horizontal angles: 'zenith,azimuth', 'azimuth,zenith',
            'elevation,azimuth' and 'azimuth,elevation'.
        conventionout : str
            The spherical coordinate convention for the output equatorial
            angles.
        degrees : boolean, optional
            If true, the angle units are degrees (radians otherwise).

        """
        operands = [
            Cartesian2SphericalOperator(conventionout, degrees=degrees),
            CartesianHorizontal2EquatorialOperator(
                convention_horizontal, time, latitude, longitude
            ),
            Spherical2CartesianOperator(conventionin, degrees=degrees),
        ]
        CompositionOperator.__init__(self, operands, **keywords)



import numpy as np
from astropy.time import Time, TimeDelta

import pyoperators
from pyoperators import (
    MPI,
    DifferenceOperator,
    NormalizeOperator,
    Spherical2CartesianOperator,
)
from pyoperators.utils import deprecated, isscalarlike, tointtuple
from pyoperators.utils.mpi import as_mpi

from ..datatypes import Map
from ..operators import (
    SphericalEquatorial2GalacticOperator,
    SphericalHorizontal2EquatorialOperator,
)
from ..quantities import Quantity
from .core import PackedTable

__all__ = ['Sampling', 'SamplingSpherical', 'SamplingEquatorial', 'SamplingHorizontal']


class Sampling(PackedTable):
    DEFAULT_DATE_OBS = '2000-01-01'
    DEFAULT_PERIOD = 1

    def __init__(self, n, date_obs=None, period=None, **keywords):
        if date_obs is None:
            date_obs = self.DEFAULT_DATE_OBS
        if isinstance(date_obs, str):
            # XXX astropy.time bug needs []
            date_obs = Time([date_obs], scale='utc')
        elif not isinstance(date_obs, Time):
            raise TypeError('The observation start date is invalid.')
        elif date_obs.is_scalar:  # work around astropy.time bug
            date_obs = Time([str(date_obs)], scale='utc')
        if period is None:
            if hasattr(keywords, 'time'):
                period = np.median(np.diff(keywords['time']))
            else:
                period = self.DEFAULT_PERIOD
        masked = keywords.pop('masked', False)
        time = keywords.pop('time', None)
        PackedTable.__init__(self, n, masked=masked, time=time, **keywords)
        self.date_obs = date_obs
        self.period = float(period)

    def time(self):
        return self.index * self.period


class SamplingSpherical(Sampling):
    def __init__(self, *args, **keywords):
        """
        ptg = SamplingSpherical(n)
        ptg = SamplingSpherical(precession=, nutation=, intrinsic_rotation=)
        ptg = SamplingSpherical(precession, nutation, intrinsic_rotation)

        Parameters
        ----------
        n : integer
            The number of samples.
        names : tuple of 3 strings
            The names of the 3 rotation angles. They are stored as Layout special
            attributes. By default, these are the proper Euler angles: precession,
            nutation and intrinsic rotation.
        degrees : boolean, optional
            If true, the input spherical coordinates are assumed to be in
            degrees.
        """
        names = keywords.pop('names', ('precession', 'nutation', 'intrinsic_rotation'))
        degrees = keywords.pop('degrees', False)
        if len(names) != 3:
            raise ValueError('The 3 pointing angles are not named.')

        if len(args) <= 1:
            if (
                names[0] in keywords
                and names[1] not in keywords
                or names[1] in keywords
                and names[0] not in keywords
            ):
                raise ValueError('The pointing is not specified.')
            if names[0] not in keywords:
                keywords[names[0]] = None
            if names[1] not in keywords:
                keywords[names[1]] = None
            if names[2] not in keywords:
                keywords[names[2]] = 0
        elif len(args) <= 3:
            keywords[names[0]] = args[0]
            keywords[names[1]] = args[1]
            keywords[names[2]] = args[2] if len(args) == 3 else 0
        else:
            raise ValueError('Invalid number of arguments.')

        if len(args) == 1:
            if not isscalarlike(args[0]):
                raise ValueError('Invalid number of arguments.')
            shape = tointtuple(args[0])
        else:
            shape = np.broadcast(
                *([keywords[_] for _ in names] + [keywords.get('time', None)])
            ).shape
        if len(shape) == 0:
            shape = (1,)
        elif len(shape) != 1:
            raise ValueError('Invalid dimension for the pointing.')
        Sampling.__init__(
            self, shape, cartesian=None, spherical=None, velocity=None, **keywords
        )
        self.names = names
        self.degrees = bool(degrees)

    def cartesian(self):
        """
        Return the cartesian coordinates in the equatorial referential.

        """
        return self.spherical2cartesian(self.spherical)

    def spherical(self):
        """
        Return the spherical coordinates as an (N, 2) ndarray.

        """
        return np.array([getattr(self, self.names[0]), getattr(self, self.names[1])]).T

    @property
    def spherical2cartesian(self):
        """
        Return the spherical-to-cartesian transform.

        """
        raise NotImplementedError()

    @property
    def cartesian2spherical(self):
        """
        Return the cartesian-to-spherical transform.

        """
        return self.spherical2cartesian.I

    def velocity(self):
        op = NormalizeOperator() * DifferenceOperator(axis=-2) / self.period
        velocity = Quantity(op(self.cartesian), 'rad/s')
        return velocity.tounit(self.DEFAULT_VELOCITY_UNIT)

    def get_center(self):
        """
        Return the average pointing direction.

        """
        if isscalarlike(self.masked) and self.masked:
            n = 0
            coords = np.zeros((1, 3))
        else:
            coords = self.cartesian
            if not isscalarlike(self.masked):
                valid = 1 - self.masked
                coords *= valid[:, None]
                n = np.sum(valid)
            else:
                n = len(self)
        n = np.asarray(n)
        center = np.sum(coords, axis=0)
        self.comm.Allreduce(MPI.IN_PLACE, as_mpi(n))
        self.comm.Allreduce(MPI.IN_PLACE, as_mpi(center))
        if n == 0:
            raise ValueError('There is no valid pointing.')
        center /= n
        return self.cartesian2spherical(center)


class SamplingEquatorial(SamplingSpherical):
    def __init__(self, *args, **keywords):
        """
        ptg = SamplingEquatorial(n)
        ptg = SamplingEquatorial(ra, dec[, pa])
        ptg = SamplingEquatorial(ra=..., dec=...[, pa=...])

        Parameters
        ----------
        n : integer
            The number of samples.
        ra : array-like
            The Right Ascension, in degrees.
        dec : array-like
            The declination, in degrees.
        pa : array-like, optional
            The position angle, in degrees (by default: 0).
        """
        SamplingSpherical.__init__(
            self,
            degrees=True,
            names=('ra', 'dec', 'pa'),
            galactic=None,
            *args,
            **keywords,
        )

    @property
    def spherical2cartesian(self):
        return Spherical2CartesianOperator('azimuth,elevation', degrees=self.degrees)

    @property
    def galactic(self):
        """
        Return the spherical coordinates in the galactic referential.

        """
        e2g = SphericalEquatorial2GalacticOperator(degrees=True)
        return e2g(self.spherical)

    def plot(
        self,
        map=None,
        header=None,
        title=None,
        new_figure=True,
        linewidth=2,
        percentile=0.01,
        **kw,
    ):
        """
        Plot the pointings' celestial coordinates.

        Parameters
        ----------
        map : ndarray of dim 2
            An optional map to be displayed as background.
        header : pyfits.Header
            The optional map's FITS header.
        title : str
            The figure's title.
        new_figure : boolean
            If true, the display will be done in a new window.
        linewidth : float
            The scan line width.
        percentile : float, tuple of two floats
            As a float, percentile of values to be discarded, otherwise,
            percentile of the minimum and maximum values to be displayed.

        """
        import kapteyn.maputils as km
        import matplotlib.pyplot as mp

        if isscalarlike(self.masked) and self.masked or np.all(self.masked):
            if new_figure:
                raise ValueError('There is no valid coordinates.')
            return
        if not isscalarlike(self.masked):
            ra = self.ra.copy()
            dec = self.dec.copy()
            ra[self.masked] = np.nan
            dec[self.masked] = np.nan
        else:
            ra = self.ra
            dec = self.dec

        if header is None:
            header = getattr(map, 'header', None)
        elif map is not None:
            map = Map(map, header=header, copy=False)

        if isinstance(map, Map) and map.has_wcs():
            image = map.imshow(
                title=title, new_figure=new_figure, percentile=percentile, **kw
            )
        else:
            if header is None:
                header = self.get_map_header(naxis=1)
            fitsobj = km.FITSimage(externalheader=dict(header))
            if new_figure:
                fig = mp.figure()
                frame = fig.add_axes((0.1, 0.1, 0.8, 0.8))
            else:
                frame = mp.gca()
            if title is not None:
                frame.set_title(title)
            image = fitsobj.Annotatedimage(frame, blankcolor='w')
            image.Graticule()
            image.plot()
            image.interact_toolbarinfo()
            image.interact_writepos()

        mp.show()
        _plot_scan(image, ra, dec, linewidth=linewidth, **kw)
        return image


class SamplingHorizontal(SamplingSpherical):
    DEFAULT_LATITUDE = None
    DEFAULT_LONGITUDE = None

    def __init__(self, *args, **keywords):
        """
        ptg = SamplingHorizontal(n, latitude=45, longitude=30)
        ptg = SamplingHorizontal(azimuth, elevation[, pitch],
                                 latitude=45, longitude=30)
        ptg = SamplingHorizontal(n, azimuth=..., elevation=..., pitch=...
                                 latitude=45, longitude=30)

        Parameters
        ----------
        n : integer
            The number of samples.
        azimuth : array-like
            The azimuth, in degrees.
        elevation : array-like
            The elevation, in degrees.
        pitch : array-like, optional
            The pitch angle, in degrees (by default: 0).

        """
        latitude = keywords.pop('latitude', None)
        if latitude is None:
            latitude = self.DEFAULT_LATITUDE
            if latitude is None:
                raise ValueError('The reference latitude is not specified.')
        self.latitude = latitude

        longitude = keywords.pop('longitude', None)
        if longitude is None:
            longitude = self.DEFAULT_LONGITUDE
            if longitude is None:
                raise ValueError('The reference longitude is not specified.')
        self.longitude = longitude

        SamplingSpherical.__init__(
            self,
            degrees=True,
            names=('azimuth', 'elevation', 'pitch'),
            equatorial=None,
            galactic=None,
            *args,
            **keywords,
        )

    @property
    def spherical2cartesian(self):
        return Spherical2CartesianOperator('azimuth,elevation', degrees=self.degrees)

    @property
    def equatorial(self):
        """
        Return the spherical coordinates in the equatorial referential.

        """
        time = self.date_obs + TimeDelta(self.time, format='sec')
        h2e = SphericalHorizontal2EquatorialOperator(
            'NE', time, self.latitude, self.longitude, degrees=True
        )
        return h2e(self.spherical, preserve_input=False)

    @property
    def galactic(self):
        """
        Return the spherical coordinates in the galactic referential.

        """
        time = self.date_obs + TimeDelta(self.time, format='sec')
        h2e = SphericalHorizontal2EquatorialOperator(
            'NE', time, self.latitude, self.longitude, degrees=True
        )
        e2g = SphericalEquatorial2GalacticOperator(degrees=True)
        return e2g(h2e)(self.spherical, preserve_input=False)


class _SamplingEquatorialScan(SamplingEquatorial):
    INSCAN = 0
    TURNAROUND = 1
    OTHER = 2

    def __init__(self, *args, **keywords):
        SamplingEquatorial.__init__(
            self, info=keywords.pop('info', 0), *args, **keywords
        )


def _plot_scan(image, ra, dec, linewidth=None, **kw):
    import matplotlib.pyplot as mp

    x, y = image.topixel(ra, dec)
    p = mp.plot(x, y, linewidth=linewidth, **kw)
    for x_, y_ in np.broadcast(x, y):
        if np.isfinite(x_) and np.isfinite(y_):
            mp.plot(x_, y_, 'o', color=p[0]._color)
            break


if pyoperators.__version__ >= '0.13.6':

    @deprecated
    class PointingSpherical(SamplingSpherical):
        pass

    @deprecated
    class PointingEquatorial(SamplingEquatorial):
        pass

    @deprecated
    class PointingHorizontal(SamplingHorizontal):
        pass

else:
    PointingSpherical = SamplingSpherical
    PointingEquatorial = SamplingEquatorial
    PointingHorizontal = SamplingHorizontal



import warnings


class PySimulatorsWarning(UserWarning):
    pass


class PySimulatorsDeprecationWarning(DeprecationWarning):
    pass


warnings.simplefilter('always', category=PySimulatorsWarning)
warnings.simplefilter('module', category=PySimulatorsDeprecationWarning)
del warnings



from .operators import (
    Cartesian2HealpixOperator,
    Healpix2CartesianOperator,
    Healpix2SphericalOperator,
    HealpixConvolutionGaussianOperator,
    HealpixLaplacianOperator,
    Spherical2HealpixOperator,
)
from .scenes import SceneHealpix, SceneHealpixCMB

__all__ = [
    'Healpix2CartesianOperator',
    'Cartesian2HealpixOperator',
    'Healpix2SphericalOperator',
    'Spherical2HealpixOperator',
    'HealpixConvolutionGaussianOperator',
    'HealpixLaplacianOperator',
    'SceneHealpix',
    'SceneHealpixCMB',
]



import numpy as np
from scipy.constants import c, h, k

from pyoperators import ReciprocalOperator, asoperator
from pyoperators.utils import strenum

from ...packedtables import Scene
from .operators import Spherical2HealpixOperator

__all__ = ['SceneHealpix', 'SceneHealpixCMB']


class SceneHealpix(Scene):
    """
    Class for Healpix scenes.

    """

    def __init__(
        self, nside, kind='I', nest=False, convention='zenith,azimuth', **keywords
    ):
        """
        nside : int
            Value of the map resolution parameter.
        kind : 'I', 'QU' or 'IQU'
            The kind of sky: intensity-only, Q and U Stokes parameters only or
            intensity, Q and U.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            the ring scheme.
        convention : string, optional
            Convention used by the 'topixel' operator to convert spherical
            coordinates into Healpix pixel numbers. It can be:
            'zenith,azimuth', 'azimuth,zenith', 'elevation,azimuth' and
            'azimuth,elevation'.

        """
        kinds = 'I', 'QU', 'IQU'
        if not isinstance(kind, str):
            raise TypeError(
                f'Invalid type {type(kind).__name__!r} for the scene kind. Expected '
                f'type is string.'
            )
        kind = kind.upper()
        if kind not in kinds:
            raise ValueError(
                f'Invalid scene kind {kind!r}. Expected kinds are: {strenum(kinds)}.'
            )
        nside = int(nside)
        topixel = Spherical2HealpixOperator(nside, convention, nest)
        shape = (12 * nside**2,)
        if kind != 'I':
            shape += (len(kind),)
        Scene.__init__(self, shape, topixel, ndim=1, **keywords)
        self.nside = nside
        self.npixel = 12 * nside**2
        self.kind = kind
        self.convention = convention
        self.nest = bool(nest)
        self.solid_angle = 4 * np.pi / self.npixel


class SceneHealpixCMB(SceneHealpix):
    def __init__(self, nside, kind='I', absolute=False, temperature=2.7255, **keywords):
        """
        Parameters
        ----------
        nside : int
            The Healpix scene's nside.
        kind : 'I', 'QU' or 'IQU', optional
            The sky kind: 'I' for intensity-only, 'QU' for Q and U maps,
            and 'IQU' for intensity plus QU maps.
        nest : boolean, optional
            For the nested numbering scheme, set it to True. Default is
            the ring scheme.
        absolute : boolean, optional
            If true, the scene pixel values include the CMB background and the
            fluctuations in units of Kelvin, otherwise it only represents the
            fluctuations, in microKelvin.
        temperature : float, optional
            The CMB black-body temperature which is used (if absolute is False)
            to convert a temperature fluctuation into power fluctuations.
            The default value is taken from Fixsen et al. 2009.

        """
        self.absolute = bool(absolute)
        self.temperature = float(temperature)
        SceneHealpix.__init__(self, nside, kind=kind, **keywords)

    def get_unit_conversion_operator(self, nu):
        """
        Return an operator to convert sky temperatures into
        W / m^2 / Hz / pixel.

        If the scene has been initialized with the 'absolute' keyword, the
        scene is assumed to include the CMB background and the fluctuations
        (in Kelvin) and the operator follows the non-linear Planck law.
        Otherwise, the scene only includes the fluctuations (in microKelvin)
        and the operator is linear (i.e. the output also corresponds to power
        fluctuations).

        Parameter
        ---------
        nu : float
            The frequency, at which the conversion is performed [Hz].

        Example
        -------
        > scene = SceneHealpixCMB(256, absolute=False)
        > op = scene.get_unit_conversion_operator(150e9)
        > dT = 200  # µK
        > print(op(dT))  # W / m^2 / Hz / pixel
        1.2734621598076659e-26

        > scene = SceneHealpixCMB(256, absolute=True)
        > op = scene.get_unit_conversion_operator(150e9)
        > T = 2.7  # K
        > print(op(T))  # W / m^2 / Hz / pixel
        5.94046610468e-23

        """
        a = 2 * self.solid_angle * h * nu**3 / c**2
        if self.absolute:
            hnu_k = h * nu / k
            return a / asoperator(np.expm1)(hnu_k * ReciprocalOperator())
        T = self.temperature
        hnu_kT = h * nu / (k * T)
        val = 1e-6 * a * hnu_kT * np.exp(hnu_kT) / (np.expm1(hnu_kT) ** 2 * T)
        return asoperator(val)



import pickle

import astropy.io.fits as fits
import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_equal

from pysimulators import FitsArray, Map, Quantity, Tod, create_fitsheader

TYPES = [Quantity, FitsArray, Map, Tod]
KEYWORDS_Q = {'unit': 'km', 'derived_units': {'km': Quantity(1000, 'm')}}
KEYWORDS_F = KEYWORDS_Q.copy()
KEYWORDS_F.update({'header': fits.Header({'NAXIS': 0})})
KEYWORDS_M = KEYWORDS_F.copy()
KEYWORDS_M.update({'coverage': 8, 'error': 1.0, 'origin': 'upper'})
KEYWORDS_T = KEYWORDS_F.copy()
KEYWORDS_T.update({'mask': True})
KEYWORDS = [KEYWORDS_Q, KEYWORDS_F, KEYWORDS_M, KEYWORDS_T]
DTYPES = [bool, int, np.float32, np.float64, np.complex64, np.complex128]


def assert_datatype_allclose(d1, d2) -> None:
    assert isinstance(d2, Quantity)
    cls = type(d2)
    assert type(d1) is cls
    assert_allclose(d1.magnitude, d2.magnitude)

    assert d1._unit == d2._unit
    assert d1._derived_units == d2._derived_units

    if cls is FitsArray:
        assert d1._header == d2._header

    elif cls is Map:
        if d2.coverage is None:
            assert d1.coverage is None
        else:
            assert d1.coverage.shape == d2.coverage.shape
            assert_equal(d1.coverage, d2.coverage)
        if d2.error is None:
            assert d1.error is None
        else:
            assert d1.error.shape == d2.error.shape
            assert_equal(d1.error, d2.error)

    elif cls is Tod:
        if d2.mask is None:
            assert d1.mask is None
        else:
            assert d1.mask.shape == d2.mask.shape
            assert_equal(d1.mask, d2.mask)


def get_datatypes_objects():
    a = np.ones((4, 3))
    a[1, 2] = 4

    q = Quantity(a, unit='myunit', derived_units={'myunit': Quantity(2.0, 'Jy')})

    header = create_fitsheader(fromdata=q, cdelt=0.5, crval=(4.0, 8.0))
    header['BUNIT'] = 'myunit'
    f = FitsArray(q, header=header)

    m = Map(f, origin='upper', error=a * 2, coverage=a * 3)

    mask = np.zeros((4, 3), np.bool_)
    mask[0, 2] = True
    t = Tod(f, mask=mask)
    return q, f, m, t


@pytest.mark.parametrize('obj', get_datatypes_objects())
@pytest.mark.parametrize('type_', TYPES)
def test_copy_false_subok_true(obj, type_):
    other = type_(obj, copy=False, subok=True)
    if isinstance(obj, type_):
        assert obj is other
    else:
        assert obj is not other
        assert_equal(obj, other)


@pytest.mark.parametrize('obj', get_datatypes_objects())
@pytest.mark.parametrize('type_', TYPES)
def test_copy_false_subok_false(obj, type_):
    other = type_(obj, copy=False, subok=False)
    if type(obj) is type_:
        assert obj is other
    else:
        assert obj is not other
        assert_equal(obj, other)


@pytest.mark.parametrize('obj', get_datatypes_objects())
@pytest.mark.parametrize('type_', TYPES)
def test_copy_true_subok_true(obj, type_):
    other = type_(obj, copy=True, subok=True)
    assert obj is not other
    assert_equal(obj, other)
    if isinstance(obj, type_):
        assert type(other) is type(obj)
    else:
        assert type(other) is type_


@pytest.mark.parametrize('obj', get_datatypes_objects())
@pytest.mark.parametrize('type_', TYPES)
def test_copy_true_subok_false(obj, type_):
    other = type_(obj, copy=True, subok=False)
    assert obj is not other
    assert_equal(obj, other)
    assert type(other) is type_


@pytest.mark.parametrize('type_', TYPES)
@pytest.mark.parametrize('input', [(2,), [2], np.array([2])])
def test_input1(type_, input):
    d = type_(input)
    assert d.shape == (1,)
    assert d[0] == 2


@pytest.mark.parametrize('type_', TYPES)
@pytest.mark.parametrize('input', [2, np.array(2)])
def test_input2(type_, input):
    d = type_(input)
    assert d.shape == ()
    assert d[...] == 2


@pytest.mark.parametrize('type_', TYPES)
def test_input3(type_):
    d = type_([])
    assert d.shape == (0,)


@pytest.mark.parametrize(
    'type_, attrs',
    [
        (Quantity, ['unit', 'derived_units']),
        (FitsArray, ['unit', 'derived_units', 'header']),
        (Map, ['unit', 'derived_units', 'header', 'coverage', 'error', 'origin']),
        (Tod, ['unit', 'derived_units', 'header', 'mask']),
    ],
)
def test_view(type_, attrs):

    data = np.ones((2, 4)).view(type_)
    for attr in attrs:
        assert hasattr(data, attr)


@pytest.mark.parametrize('type_', TYPES)
def test_operation(type_):
    a = type_.ones(10)
    assert type(a + 1) is type_


@pytest.mark.parametrize('type_', TYPES)
@pytest.mark.parametrize('dtype', DTYPES)
def test_dtype1(type_, dtype):
    a = type_(np.array([10, 20], dtype=dtype))
    assert a.dtype == (dtype if dtype != int else float)


@pytest.mark.parametrize('type_', TYPES)
@pytest.mark.parametrize('dtype', DTYPES)
def test_dtype2(type_, dtype):
    a = type_(np.array([10, 20], dtype=dtype), dtype=dtype)
    assert a.dtype == dtype


@pytest.mark.parametrize('type_', TYPES)
@pytest.mark.parametrize('dtype', DTYPES)
def test_dtype3(type_, dtype):
    a = type_(np.ones((4, 3)), dtype=dtype, copy=None)
    assert a.dtype == dtype


@pytest.mark.parametrize('type_, keywords', zip(TYPES, KEYWORDS))
@pytest.mark.parametrize('method', ['empty', 'ones', 'zeros'])
@pytest.mark.parametrize('dtype', [None, *DTYPES])
def test_empty_ones_zeros(type_, keywords, method, dtype):
    method = getattr(type_, method)
    a = method((2, 3), dtype=dtype, **keywords)
    assert a.dtype == type_.default_dtype if dtype is None else dtype
    for k_, v in keywords.items():
        assert getattr(a, k_) == v


def test_map():
    class MAP(Map):
        def __init__(self, data):
            self.info = 'info'

    m = MAP(np.ones(3))
    assert sorted(m.__dict__) == [
        '_derived_units',
        '_header',
        '_unit',
        'coverage',
        'error',
        'info',
        'origin',
    ]


@pytest.mark.parametrize('obj', get_datatypes_objects())
@pytest.mark.parametrize('protocol', range(pickle.HIGHEST_PROTOCOL))
def test_pickling(obj, protocol):
    actual = pickle.loads(pickle.dumps(obj, protocol))
    assert_equal(obj, actual)


@pytest.mark.parametrize('obj', get_datatypes_objects())
def test_save(tmp_path, obj):
    if not isinstance(obj, FitsArray):
        assert not hasattr(obj, 'save')
        return

    filename = str(tmp_path / 'obj.fits')
    obj.save(filename)
    actual = type(obj)(filename)
    assert_equal(actual, obj)


@pytest.mark.parametrize('cls', [Quantity, FitsArray, Map, Tod])
@pytest.mark.parametrize(
    'func, expected_unit',
    [
        (np.min, {'u': 1}),
        (np.max, {'u': 1}),
        (np.sum, {'u': 1}),
        (np.mean, {'u': 1}),
        (np.ptp, {'u': 1}),
        (np.std, {'u': 1}),
        (np.var, {'u': 2}),
    ],
)
@pytest.mark.parametrize('axis', [None, 0, 1])
def test_function_reduction(cls, func, expected_unit, axis):
    data = [[1.0, 2, 3, 4], [5, 6, 7, 8]]
    array = cls(data, unit='u')

    actual = func(array, axis=axis)

    np_output = func(data, axis=axis)
    expected = cls(np_output, unit=expected_unit)
    assert_datatype_allclose(actual, expected)


@pytest.mark.parametrize(
    'func, expected_unit',
    [
        (np.min, {'u': 1}),
        (np.max, {'u': 1}),
        (np.sum, {'u': 1}),
        (np.mean, {'u': 1}),
        (np.ptp, {'u': 1}),
        (np.std, {'u': 1}),
        (np.var, {'u': 2}),
    ],
)
@pytest.mark.parametrize('axis', [None, 0, 1])
def test_function_reduction_tod(func, expected_unit, axis):
    data = [[1.0, 2, 3, 4], [5, 6, 7, 8]]
    mask = [[True, False, False, False], [False, True, True, False]]
    array = Tod(data, mask=mask, unit='u')

    actual = func(array, axis=axis)

    ma_output = func(np.ma.MaskedArray(data, mask=mask), axis=axis)
    if np.isscalar(ma_output):
        expected = Tod(ma_output, mask=None, unit=expected_unit)
    else:
        expected = Tod(ma_output.data, mask=ma_output.mask, unit=expected_unit)
    assert_datatype_allclose(actual, expected)


@pytest.mark.parametrize('cls', [Quantity, FitsArray, Map, Tod])
@pytest.mark.parametrize(
    'func, expected_unit',
    [
        (np.round, {'u': 1}),
        (np.exp, {}),
        (np.square, {'u': 2}),
        (np.reciprocal, {'u': -1}),
    ],
)
def test_ufunc(cls, func, expected_unit):
    array = cls([[1.0, 2, 3, 4], [5, 6, 7, 8]], unit='u')

    actual = func(array)

    output = func(array.view(np.ndarray))
    expected = cls(output, unit=expected_unit)
    assert_datatype_allclose(actual, expected)


@pytest.mark.parametrize(
    'func, expected_unit',
    [
        (np.round, {'u': 1}),
        (np.exp, {}),
        (np.square, {'u': 2}),
        (np.reciprocal, {'u': -1}),
    ],
)
def test_ufunc_tod(func, expected_unit):
    data = [[1, 2, 3, 4], [5, 6, 7, 8]]
    mask = [[True, False, False, False], [False, True, True, False]]
    array = Tod(data, mask=mask, unit='u')

    actual = func(array)

    output = func(array.view(np.ndarray))
    expected = Tod(output, mask=mask, unit=expected_unit)
    assert_datatype_allclose(actual, expected)


@pytest.mark.parametrize(
    'array',
    [
        Quantity([[1.0, 2]], unit='u'),
        FitsArray([[1.0, 2]], unit='u'),
        Map([[1.0, 2]], unit='u'),
        Map([[1.0, 2]], coverage=[[0.0, 1]], error=[[0.5, 1]], unit='u'),
        Tod([[1.0, 2]], unit='u'),
        Tod([[1.0, 2]], mask=[[True, False]], unit='u'),
    ],
)
@pytest.mark.parametrize(
    'func, new_shape',
    [
        (lambda a: np.ravel(a), (2,)),
        (lambda a: a.ravel(), (2,)),
        (lambda a: np.reshape(a, (2, 1)), (2, 1)),
        (lambda a: a.reshape(2, 1), (2, 1)),
        (lambda a: a.reshape((2, 1)), (2, 1)),
    ],
)
def test_ravel_reshape(array, func, new_shape):
    reshaped_array = func(array)

    assert type(reshaped_array) is type(array)
    assert reshaped_array.shape == new_shape
    if isinstance(array, Map):
        if array.coverage is None:
            assert reshaped_array.coverage is None
        else:
            assert reshaped_array.coverage.shape == new_shape
        if array.error is None:
            assert reshaped_array.error is None
        else:
            assert reshaped_array.error.shape == new_shape
    elif isinstance(array, Tod):
        if array.mask is None:
            assert reshaped_array.mask is None
        else:
            assert reshaped_array.mask.shape == new_shape


@pytest.mark.parametrize(
    'dtype1', [np.int8, np.int32, np.int64, np.float32, np.float64]
)
@pytest.mark.parametrize(
    'dtype2', [np.int8, np.int32, np.int64, np.float32, np.float64]
)
def test_astype(dtype1, dtype2):
    m = Map(
        np.array([1, 2, 3], dtype=dtype1),
        coverage=np.array([0, 1, 0], dtype=dtype1),
        error=np.array([2, 2, 2], dtype=dtype1),
    )
    m2 = m.astype(dtype2)
    assert m2.dtype == dtype2
    assert m2.coverage.dtype == dtype2
    assert m2.error.dtype == dtype2



# Copyrights 2010-2011 Pierre Chanial
# All rights reserved
#
"""
The datatypes module contains 3 ndarray subclasses:
    - FitsArray
    - Map
    - Tod
The Map and Tod classes subclass FitsArray.

These classes are useful to load, manipulate and save FITs files.
They also contain specialised display methods.

"""

import io
import os
import pickle
import sys
import time
import uuid
from functools import reduce
from pathlib import Path

import numpy as np
import scipy.stats
from astropy.io import fits as pyfits

from pyoperators import MPI
from pyoperators.memory import empty
from pyoperators.utils import isscalarlike

from .mpiutils import read_fits, write_fits
from .quantities import Quantity
from .wcsutils import create_fitsheader_for, has_wcs

__all__ = ['FitsArray', 'Map', 'Tod']


class FitsArray(Quantity):
    """
    FitsArray(filename|object, header=None, unit=None, derived_units=None,
              dtype=None, copy=True, order='C', subok=False, ndmin=0,
              comm=None)

    An ndarray subclass, whose instances
        - store the FITS header information
        - can be read from and written to FITS files
        - have a specialised matplotlib display and export
        - can be easily displayed using ds9

    Parameters
    ----------
    filename : str | Path
        The FITS file name.
    object : array_like
        An array, any object exposing the array interface, an
        object whose __array__ method returns an array, or any
        (nested) sequence.
    header : pyfits.Header, optional
        The FITS header.
    unit : str, optional
        The data unit.
    derived_units : dict, optional
        Dictionary defining non-standard units.
    dtype : data-type, optional
        The desired data-type for the array.  If not given, then
        the type will be determined as the minimum type required
        to hold the objects in the sequence.  This argument can only
        be used to 'upcast' the array.  For downcasting, use the
        .astype(t) method.
    copy : bool, optional
        If true (default), then the object is copied.  Otherwise, a copy
        will only be made if __array__ returns a copy, if obj is a
        nested sequence, or if a copy is needed to satisfy any of the other
        requirements (`dtype`, `order`, etc.).
    order : {'C', 'F', 'A'}, optional
        Specify the order of the array.  If order is 'C' (default), then the
        array will be in C-contiguous order (last-index varies the
        fastest).  If order is 'F', then the returned array
        will be in Fortran-contiguous order (first-index varies the
        fastest).  If order is 'A', then the returned array may
        be in any order (either C-, Fortran-contiguous, or even
        discontiguous).
    subok : bool, optional
        If True, then sub-classes will be passed-through, otherwise
        the returned array will be forced to be a base-class array (default).
    ndmin : int, optional
        Specifies the minimum number of dimensions that the resulting
        array should have.  Ones will be pre-pended to the shape as
        needed to meet this requirement.
    comm : mpi4py.Comm, optional
        MPI communicator specifying to which processors the FITS file
        should be distributed.

    """

    _header = None

    def __new__(
        cls,
        data,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        copy=True,
        order='C',
        subok=False,
        ndmin=0,
        comm=None,
    ):

        if isinstance(data, (str, Path)):

            if comm is None:
                comm = MPI.COMM_SELF

            try:
                buf = pyfits.open(data)['derived_units'].data
                derived_units = pickle.loads(buf.data)
            except (ImportError, KeyError):
                pass

            data, header = read_fits(data, None, comm)

            copy = False
            if unit is None:
                if 'BUNIT' in header:
                    unit = header['BUNIT']
                elif 'QTTY____' in header:
                    unit = header['QTTY____']  # HCSS crap
                    if unit == '1':
                        unit = ''

        elif comm is not None:
            raise ValueError(
                'The MPI communicator can only be set for input F' 'ITS files.'
            )

        # get a new FitsArray instance (or a subclass if subok is True)
        result = Quantity.__new__(
            cls, data, unit, derived_units, dtype, copy, order, True, ndmin
        )
        if not subok and type(result) is not cls:
            result = result.view(cls)

        # copy header attribute
        if header is not None:
            result.header = header
        elif hasattr(data, '_header') and type(data._header) is pyfits.Header:
            if copy:
                result._header = data._header.copy()
            else:
                result._header = data._header
        else:
            result._header = None

        return result

    def __array_finalize__(self, array):
        super().__array_finalize__(array)
        self._header = getattr(array, '_header', None)

    def __getattr__(self, name):
        if self.dtype.names and name in self.dtype.names:
            return self[name]
        return super().__getattribute__(name)

    def __setattr__(self, name, value):
        if self.dtype.names and name in self.dtype.names:
            self[name] = value
        else:
            super().__setattr__(name, value)

    def astype(self, dtype):
        """
        f.astype(t)

        Copy of the FitsArray, cast to a specified type. The FITS keyword
        BITPIX is updated accordingly.

        Parameters
        ----------
        t : str or dtype
            Typecode or data-type to which the array is cast.

        Raises
        ------
        ComplexWarning :
            When casting from complex to float or int. To avoid this,
            one should use ``m.real.astype(t)``.

        Examples
        --------
        >>> f = FitsArray([1, 2, 2.5])
        >>> f.astype(int)
        FitsArray([1 2 2], '')

        """
        result = super().astype(dtype)
        if self._header is not None:
            typename = np.dtype(dtype).name
            result._header = self._header.copy()
            result._header['BITPIX'] = pyfits.DTYPE2BITPIX[typename]
        return result

    @classmethod
    def empty(
        cls,
        shape,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            empty(shape, dtype, order),
            header=header,
            unit=unit,
            derived_units=derived_units,
            dtype=dtype,
            copy=False,
            **keywords,
        )

    @classmethod
    def ones(
        cls,
        shape,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.ones(shape, dtype, order),
            header=header,
            unit=unit,
            derived_units=derived_units,
            dtype=dtype,
            copy=False,
            **keywords,
        )

    @classmethod
    def zeros(
        cls,
        shape,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.zeros(shape, dtype, order),
            header=header,
            unit=unit,
            derived_units=derived_units,
            dtype=dtype,
            copy=False,
            **keywords,
        )

    def has_wcs(self):
        """
        Returns True is the array has a FITS header with a defined World
        Coordinate System.
        """
        return has_wcs(self.header)

    @property
    def header(self):
        if self._header is None and not np.iscomplexobj(self):
            self._header = create_fitsheader_for(self)
        return self._header

    @header.setter
    def header(self, header):
        if header is not None and not isinstance(header, pyfits.Header):
            raise TypeError(
                'Incorrect type for the input header (' + str(type(header)) + ').'
            )
        self._header = header

    def save(self, filename, comm=None):
        """
        Write array as a FITS file.

        Parameters
        ----------
        filename : str
            The FITS file name.
        comm : mpi4py.Comm, optional
            MPI communicator specifying which processors should write
            to the FITS file.
        """

        header = self.header.copy()
        if comm is None:
            comm = MPI.COMM_SELF

        if len(self._unit) != 0:
            header['BUNIT'] = self.unit

        write_fits(filename, self, header, False, None, comm)
        if not self.derived_units:
            return
        if comm is None or comm.rank == 0:
            buf = io.BytesIO()
            pickle.dump(self.derived_units, buf, pickle.HIGHEST_PROTOCOL)
            data = np.frombuffer(buf.getvalue(), np.uint8)
            write_fits(filename, data, None, True, 'derived_units', MPI.COMM_SELF)
        comm.Barrier()

    def imsave(self, filename, colorbar=True, **kw):
        """
        Export to file (JPG, PNG, etc.) the display that would be produced
        by the imshow method.

        Example
        -------
        >>> Map(np.arange(100).reshape((10,10))).imsave('plot.png')
        """
        import matplotlib
        import matplotlib.pyplot as mp

        is_interactive = matplotlib.is_interactive()
        matplotlib.interactive(False)
        dpi = 80.0
        figsize = np.clip(np.max(np.array(self.shape[::-1]) / dpi), 8, 50)
        figsize = (figsize + (2 if colorbar else 0), figsize)
        fig = mp.figure(figsize=figsize, dpi=dpi)
        self.imshow(colorbar=colorbar, new_figure=False, **kw)
        fig = mp.gcf()
        fig.savefig(filename)
        matplotlib.interactive(is_interactive)

    def imshow(
        self,
        mask=None,
        new_figure=True,
        title=None,
        xlabel='',
        ylabel='',
        interpolation='nearest',
        colorbar=True,
        percentile=0,
        **keywords,
    ):
        """
        A graphical display method specialising matplotlib's imshow.

        mask : array-like, optional
            Data mask. True means masked.
        new_figure : boolean, optional
            If True, a new figure if created. Default is True.
        title : str, optional
            The plot title.
        xlabel : str, optional
           Plot X label.
        ylabel : str, optional
           Plot Y label.
        interpolation : str, optional
            Acceptable values are None, 'nearest', 'bilinear',
            'bicubic', 'spline16', 'spline36', 'hanning', 'hamming',
            'hermite', 'kaiser', 'quadric', 'catrom', 'gaussian',
            'bessel', 'mitchell', 'sinc', 'lanczos'
            Default is 'nearest'.
        colorbar : boolean, optional
            If True, plot an intensity color bar next to the display.
            Default is True.
        percentile : float, tuple of two floats in range [0,100]
            As a float, percentile of values to be discarded, otherwise,
            percentile of the minimum and maximum values to be displayed.
        **keywords : additional are keywords passed to matplotlib's imshow.

        """
        import matplotlib.pyplot as mp

        if np.iscomplexobj(self):
            data = abs(self.magnitude)
        else:
            data = self.magnitude.copy()
        if isscalarlike(percentile):
            percentile = percentile / 2, 100 - percentile / 2
        unfinite = ~np.isfinite(self.magnitude)
        if mask is None:
            mask = unfinite
        else:
            mask = np.logical_or(mask, unfinite)
        data_valid = data[~mask]
        minval = scipy.stats.scoreatpercentile(data_valid, percentile[0])
        maxval = scipy.stats.scoreatpercentile(data_valid, percentile[1])
        data[data < minval] = minval
        data[data > maxval] = maxval
        data[mask] = np.nan

        if new_figure:
            fig = mp.figure()
        else:
            fig = mp.gcf()
        fontsize = 12.0 * fig.get_figheight() / 6.125

        image = mp.imshow(data, interpolation=interpolation, **keywords)
        image.set_clim(minval, maxval)

        ax = mp.gca()
        ax.set_xlabel(xlabel, fontsize=fontsize)
        ax.set_ylabel(ylabel, fontsize=fontsize)
        for tick in ax.xaxis.get_major_ticks():
            tick.label1.set_fontsize(fontsize)
        for tick in ax.yaxis.get_major_ticks():
            tick.label1.set_fontsize(fontsize)

        if title is not None:
            mp.title(title, fontsize=fontsize)
        if colorbar:
            colorbar = mp.colorbar()
            for tick in colorbar.ax.get_yticklabels():
                tick.set_fontsize(fontsize)

        mp.draw()
        return image

    def ds9(self, xpamsg=None, origin=None, new=True, **keywords):
        """
        Display the array using ds9.

        The ds9 process will be given an random id. By default, the
        following access point are set before the array is loaded:
            -cmap heat
            -scale scope local
            -scale mode 99.5
        Other access points can be set before the data is loaded though
        the keywords (see examples below).
        After the array is loaded, the map's header is set and the user
        may add other XPA messages through the xpamsg argument or by
        setting them through the returned ds9 instance.

        Parameters
        ----------
        xpamsg : string or tuple of string
            XPA access point message to be set after the array is loaded.
            (see http://hea-www.harvard.edu/RD/ds9/ref/xpa.html).
        origin: string
            Set origin to 'upper' for Y increasing downwards
        new: boolean
            If true, open the array in a new ds9 instance.
        **keywords : string or tuple of string
            Specify more access points to be set before array loading.
            a keyword such as 'height=400' will be appended to the command
            that launches ds9 in the form 'ds9 [...] -height 400'

        Returns
        -------
        The returned object is a ds9 instance. It can be manipulated using
        XPA access points.

        Examples
        --------
        >>> m = Map('myfits.fits')
        >>> d=m.ds9(('zoom to fit', 'saveimage png myfits.png'),
        ...         scale='histequ', cmap='invert yes', height=400)
        >>> d.set('exit')

        """
        try:
            import pyds9
        except ImportError:
            raise ImportError('The library pyds9 has not been installed.')
        import xpa

        id = None
        if not new:
            list = pyds9.ds9_targets()
            if list is not None:
                id = list[-1]

        if id is None:
            if 'cmap' not in keywords:
                keywords['cmap'] = 'heat'

            if 'scale' not in keywords:
                keywords['scale'] = ('scope local', 'mode 99.5')

            if origin is None:
                origin = getattr(self, 'origin', 'lower')
            if origin == 'upper' and 'orient' not in keywords:
                keywords['orient'] = 'y'

            wait = 10

            id_ = 'ds9_' + str(uuid.uuid1())[4:8]

            command = 'ds9 -title ' + id_

            for k, v in keywords.items():
                k = str(k)
                if type(v) is not tuple:
                    v = (v,)
                command += reduce(lambda x, y: str(x) + ' -' + k + ' ' + str(y), v, '')

            os.system(command + ' &')

            # start the xpans name server
            if xpa.xpaaccess('xpans', None, 1) is None:
                _cmd = None
                # look in install directories
                for _dir in sys.path:
                    _fname = os.path.join(_dir, 'xpans')
                    if os.path.exists(_fname):
                        _cmd = _fname + ' -e &'
                if _cmd:
                    os.system(_cmd)

            for i in range(wait):
                list = xpa.xpaaccess(id_, None, 1024)
                if list is not None:
                    break
                time.sleep(1)
            if not list:
                raise ValueError('No active ds9 running for target: %s' % list)

        # get ds9 instance with given id
        d = pyds9.DS9(id_)

        # load array
        input = self.view(np.ndarray)
        if input.dtype.kind in ('b', 'i'):
            input = np.asarray(input, np.int32)
        d.set_np2arr(input.T)

        # load header
        if self.has_wcs():
            d.set('wcs append', str(self.header))

        if xpamsg is not None:
            if isinstance(xpamsg, str):
                xpamsg = (xpamsg,)
            for v in xpamsg:
                d.set(v)

        return d


class Map(FitsArray):
    """
    Map(filename|object, header=None, unit=None, derived_units=None,
        coverage=None, error=None, origin='lower', dtype=None, copy=True,
        order='C', subok=False, ndmin=0, comm=None)

    A FitsArray subclass, whose instances have
        - a 'coverage' and 'error' attribute
        - a specialised imshow method, using WCS information if available
        - an attribute 'origin', which flags whether the map's origin is on
          the bottom-left ('lower') or top-left ('upper').

    Parameters
    ----------
    filename : str | Path
        The FITS file name.
    object : array_like
        An array, any object exposing the array interface, an
        object whose __array__ method returns an array, or any
        (nested) sequence.
    header : pyfits.Header, optional
        The FITS header.
    unit : str, optional
        The data unit.
    derived_units : dict, optional
        Dictionary defining non-standard units.
    coverage : ndarray, optional
        The coverage map, of the same shape as the data.
    error : ndarray, optional
        The error map, of the same shape as the data.
    origin : 'lower' or 'upper', optional
        Display convention for the Map. For an top-left origin, use 'upper'.
        For a bottom-left origin, use 'lower'. The default convention is
        bottom-left.
    dtype : data-type, optional
        The desired data-type for the array.  If not given, then
        the type will be determined as the minimum type required
        to hold the objects in the sequence.  This argument can only
        be used to 'upcast' the array.  For downcasting, use the
        .astype(t) method.
    copy : bool, optional
        If true (default), then the object is copied.  Otherwise, a copy
        will only be made if __array__ returns a copy, if obj is a
        nested sequence, or if a copy is needed to satisfy any of the other
        requirements (`dtype`, `order`, etc.).
    order : {'C', 'F', 'A'}, optional
        Specify the order of the array.  If order is 'C' (default), then the
        array will be in C-contiguous order (last-index varies the
        fastest).  If order is 'F', then the returned array
        will be in Fortran-contiguous order (first-index varies the
        fastest).  If order is 'A', then the returned array may
        be in any order (either C-, Fortran-contiguous, or even
        discontiguous).
    subok : bool, optional
        If True, then sub-classes will be passed-through, otherwise
        the returned array will be forced to be a base-class array (default).
    ndmin : int, optional
        Specifies the minimum number of dimensions that the resulting
        array should have.  Ones will be pre-pended to the shape as
        needed to meet this requirement.
    comm : mpi4py.Comm, optional
        MPI communicator specifying to which processors the FITS file
        should be distributed.

    """

    coverage = None
    error = None
    origin = None

    def __new__(
        cls,
        data,
        header=None,
        unit=None,
        derived_units=None,
        coverage=None,
        error=None,
        origin=None,
        dtype=None,
        copy=True,
        order='C',
        subok=False,
        ndmin=0,
        comm=None,
    ):

        # get a new Map instance (or a subclass if subok is True)
        result = FitsArray.__new__(
            cls,
            data,
            header,
            unit,
            derived_units,
            dtype,
            copy,
            order,
            True,
            ndmin,
            comm,
        )
        if not subok and type(result) is not cls:
            result = result.view(cls)

        if isinstance(data, str):

            if comm is None:
                comm = MPI.COMM_SELF

            if 'DISPORIG' in result.header:
                if origin is None:
                    origin = result.header['DISPORIG']
            if coverage is None:
                try:
                    coverage, junk = read_fits(data, 'Coverage', comm)
                except KeyError:
                    pass
            if error is None:
                try:
                    error, junk = read_fits(data, 'Error', comm)
                except KeyError:
                    pass

        if origin is not None:
            origin = origin.strip().lower()
            if origin not in ('lower', 'upper'):
                raise ValueError(
                    "Invalid origin '" + origin + "'. Expected va"
                    "lues are 'lower' or 'upper'."
                )
            result.origin = origin

        if coverage is not None:
            result.coverage = np.asanyarray(coverage)
        elif copy and result.coverage is not None:
            result.coverage = result.coverage.copy()

        if error is not None:
            result.error = np.asanyarray(error)
        elif copy and result.error is not None:
            result.error = result.error.copy()

        return result

    def __array_finalize__(self, array):
        super().__array_finalize__(array)
        self.coverage = getattr(array, 'coverage', None)
        self.error = getattr(array, 'error', None)
        self.origin = getattr(array, 'origin', 'lower')

    def __array_function__(self, func, types, args, kwargs):
        if func is np.reshape:
            return args[0].reshape(*args[1:], **kwargs)
        if func is np.ravel:
            return args[0].ravel(*args[1:], **kwargs)
        return super().__array_function__(func, types, args, kwargs)

    def __getitem__(self, key):
        item = super().__getitem__(key)
        if not isinstance(item, Map):
            return item
        if item.coverage is not None:
            item.coverage = item.coverage[key]
        if item.error is not None:
            item.error = item.error[key]
        return item

    def astype(self, dtype):
        """
        m.astype(t)

        Copy of the Map, cast to a specified type. The types of the coverage
        and error attribute are also converted.

        Parameters
        ----------
        t : str or dtype
            Typecode or data-type to which the array is cast.

        Raises
        ------
        ComplexWarning :
            When casting from complex to float or int. To avoid this,
            one should use ``m.real.astype(t)``.

        Examples
        --------
        >>> x = Map([1, 2, 2.5], coverage=[0., 1., 1.])
        >>> x2 = x.astype(int)
        >>> x2
        array([1, 2, 2])
        >>> x2.coverage
        array([0, 1, 1])

        """
        result = super().astype(dtype)
        if self.coverage is not None:
            result.coverage = self.coverage.astype(dtype)
        if self.error is not None:
            result.error = self.error.astype(dtype)
        return result

    @classmethod
    def empty(
        cls,
        shape,
        header=None,
        unit=None,
        derived_units=None,
        coverage=None,
        error=None,
        origin=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            empty(shape, dtype, order),
            header=header,
            unit=unit,
            derived_units=derived_units,
            coverage=coverage,
            error=error,
            origin=origin,
            copy=False,
            dtype=dtype,
            **keywords,
        )

    @classmethod
    def ones(
        cls,
        shape,
        header=None,
        unit=None,
        derived_units=None,
        coverage=None,
        error=None,
        origin=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.ones(shape, dtype, order),
            header=header,
            unit=unit,
            derived_units=derived_units,
            coverage=coverage,
            error=error,
            origin=origin,
            copy=False,
            dtype=dtype,
            **keywords,
        )

    @classmethod
    def zeros(
        cls,
        shape,
        header=None,
        unit=None,
        derived_units=None,
        coverage=None,
        error=None,
        origin=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.zeros(shape, dtype, order),
            header=header,
            unit=unit,
            derived_units=derived_units,
            coverage=coverage,
            error=error,
            origin=origin,
            copy=False,
            dtype=dtype,
            **keywords,
        )

    def imshow(
        self,
        mask=None,
        new_figure=True,
        title=None,
        xlabel='X',
        ylabel='Y',
        interpolation='nearest',
        origin=None,
        colorbar=True,
        percentile=0,
        **keywords,
    ):
        try:
            import kapteyn.maputils as km
        except ImportError:
            km = None
        import matplotlib.pyplot as mp

        if mask is None and self.coverage is not None:
            mask = self.coverage <= 0

        if origin is None:
            origin = self.origin

        # check if the map has no astrometry information
        if not self.has_wcs() or km is None:
            image = super().imshow(
                mask=mask,
                new_figure=new_figure,
                title=title,
                xlabel=xlabel,
                ylabel=ylabel,
                origin=origin,
                colorbar=colorbar,
                percentile=percentile,
                **keywords,
            )
            return image

        if np.iscomplexobj(self):
            data = abs(self.magnitude)
        else:
            data = self.magnitude.copy()
        if isscalarlike(percentile):
            percentile = percentile / 2, 100 - percentile / 2
        unfinite = ~np.isfinite(self.magnitude)
        if mask is None:
            mask = unfinite
        else:
            mask = np.logical_or(mask, unfinite)
        data_valid = data[~mask]
        minval = scipy.stats.scoreatpercentile(data_valid, percentile[0])
        maxval = scipy.stats.scoreatpercentile(data_valid, percentile[1])
        data[data < minval] = minval
        data[data > maxval] = maxval
        data[mask] = np.nan

        # XXX FIX ME
        colorbar = False

        fitsobj = km.FITSimage(externaldata=data, externalheader=dict(self.header))
        if new_figure:
            fig = mp.figure()
            frame = fig.add_axes((0.1, 0.1, 0.8, 0.8))
        else:
            frame = mp.gca()
        fontsize = 12.0 * fig.get_figheight() / 6.125
        for tick in frame.xaxis.get_major_ticks():
            tick.label1.set_fontsize(fontsize)
        for tick in frame.yaxis.get_major_ticks():
            tick.label1.set_fontsize(fontsize)
        if title is not None:
            frame.set_title(title)
        if colorbar:
            colorbar = mp.colorbar()
            for tick in colorbar.ax.get_yticklabels():
                tick.set_fontsize(fontsize)
        annim = fitsobj.Annotatedimage(frame, blankcolor='w')
        annim.Image(interpolation=interpolation)
        grat = annim.Graticule()
        grat.setp_gratline(visible=False)
        annim.plot()
        annim.interact_imagecolors()
        annim.interact_toolbarinfo()
        annim.interact_writepos()
        if colorbar:
            annim.Colorbar()
        mp.show()
        return annim

    def save(self, filename, comm=None):
        if comm is None:
            comm = MPI.COMM_SELF

        self.header['DISPORIG'] = self.origin, 'Map display convention'
        FitsArray.save(self, filename, comm=comm)
        if self.coverage is not None:
            write_fits(filename, self.coverage, None, True, 'Coverage', comm)
        if self.error is not None:
            write_fits(filename, self.error, None, True, 'Error', comm)

    def ravel(self, order='C'):
        result = super().ravel(order)
        coverage = None if self.coverage is None else self.coverage.ravel(order)
        error = None if self.error is None else self.error.ravel(order)
        result.coverage = coverage
        result.error = error
        return result

    def reshape(self, *newdims, order='C'):
        result = super().reshape(*newdims, order=order)
        coverage = (
            None
            if self.coverage is None
            else self.coverage.reshape(*newdims, order=order)
        )
        error = (
            None if self.error is None else self.error.reshape(*newdims, order=order)
        )
        result.coverage = coverage
        result.error = error
        return result


class Tod(FitsArray):
    """
    Tod(filename|object, header=None, unit=None, derived_units=None, mask=None,
        dtype=None, copy=True, order='C', subok=False, ndmin=0, comm=None)

    A FitsArray subclass for Time Ordered Data. The time is assumed to be
    the last dimension. Tod instances have
        - a 'mask' attribute, of the same shape than the data. A True value
          means that the data sample is masked
        - a specialised imshow method

    Parameters
    ----------
    filename : str | Path
        The FITS file name.
    object : array_like
        An array, any object exposing the array interface, an
        object whose __array__ method returns an array, or any
        (nested) sequence.
    header : pyfits.Header, optional
        The FITS header.
    unit : str, optional
        The data unit.
    derived_units : dict, optional
        Dictionary defining non-standard units.
    mask : boolean ndarray, optional
        The TOD mask, of the same shape as the data. True means masked.
    dtype : data-type, optional
        The desired data-type for the array.  If not given, then
        the type will be determined as the minimum type required
        to hold the objects in the sequence.  This argument can only
        be used to 'upcast' the array.  For downcasting, use the
        .astype(t) method.
    copy : bool, optional
        If true (default), then the object is copied.  Otherwise, a copy
        will only be made if __array__ returns a copy, if obj is a
        nested sequence, or if a copy is needed to satisfy any of the other
        requirements (`dtype`, `order`, etc.).
    order : {'C', 'F', 'A'}, optional
        Specify the order of the array.  If order is 'C' (default), then the
        array will be in C-contiguous order (last-index varies the
        fastest).  If order is 'F', then the returned array
        will be in Fortran-contiguous order (first-index varies the
        fastest).  If order is 'A', then the returned array may
        be in any order (either C-, Fortran-contiguous, or even
        discontiguous).
    subok : bool, optional
        If True, then sub-classes will be passed-through, otherwise
        the returned array will be forced to be a base-class array (default).
    ndmin : int, optional
        Specifies the minimum number of dimensions that the resulting
        array should have.  Ones will be pre-pended to the shape as
        needed to meet this requirement.
    comm : mpi4py.Comm, optional
        MPI communicator specifying to which processors the FITS file
        should be distributed.
    """

    _mask = None

    def __new__(
        cls,
        data,
        mask=None,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        copy=True,
        order='C',
        subok=False,
        ndmin=0,
        comm=None,
    ):

        # get a new Tod instance (or a subclass if subok is True)
        result = FitsArray.__new__(
            cls,
            data,
            header,
            unit,
            derived_units,
            dtype,
            copy,
            order,
            True,
            ndmin,
            comm,
        )
        if not subok and type(result) is not cls:
            result = result.view(cls)

        # mask attribute
        if mask is np.ma.nomask:
            mask = None

        if mask is None and isinstance(data, str):

            if comm is None:
                comm = MPI.COMM_SELF

            try:
                mask, junk = read_fits(data, 'mask', comm)
                mask = mask.view(np.bool_)
                copy = False
            except Exception:  # FIXME
                pass

        if mask is None and hasattr(data, 'mask') and data.mask is not np.ma.nomask:
            mask = data.mask

        if mask is not None:
            if not isinstance(mask, np.ndarray) or mask.dtype != bool:
                copy = True
            result._mask = np.array(mask, np.bool_, copy=copy)

        return result

    @property
    def mask(self):
        return self._mask

    @mask.setter
    def mask(self, mask):
        if mask is None or mask is np.ma.nomask:
            self._mask = None
            return

        # enforce bool dtype
        if not isinstance(mask, np.ndarray):
            mask = np.array(mask, np.bool_)
        elif mask.dtype.type != np.bool_:
            if mask.dtype.itemsize == 1:
                mask = mask.view(np.bool_)
            else:
                mask = np.asarray(mask, np.bool_)

        # handle the scalar case
        if mask.ndim == 0:
            if self._mask is None:
                func = np.zeros if mask == 0 else np.ones
                self._mask = func(self.shape, dtype=np.bool_)
            else:
                self._mask[:] = mask
            return

        # check shape compatibility
        if self.shape != mask.shape:
            raise ValueError(
                "The input mask has a shape '"
                + str(mask.shape)
                + "' incompatible with that of the Tod '"
                + str(self.shape)
                + "'."
            )

        self._mask = mask

    def __array_finalize__(self, array):
        super().__array_finalize__(array)
        self._mask = getattr(array, 'mask', None)

    def __array_function__(self, func, types, args, kwargs):
        """Handle NumPy functions on Tod objects, respecting the mask."""
        if func is np.reshape:
            return args[0].reshape(*args[1:], **kwargs)
        if func is np.ravel:
            return args[0].ravel(*args[1:], **kwargs)
        return super().__array_function__(func, types, args, kwargs)

    @classmethod
    def _view_function_input(cls, array):
        if isinstance(array, Tod) and array.mask is not None:
            return np.ma.MaskedArray(
                array.view(np.ndarray), mask=array.mask, copy=False
            )
        return super()._view_function_input(array)

    def _convert_function_outputs(self, arrays, output_units, conversion):
        """Convert function outputs back to Tod, extracting mask from MaskedArray."""
        # Handle MaskedConstant (all values masked)
        if isinstance(arrays, np.ma.core.MaskedConstant):
            output = type(self)(np.nan, mask=True)
            output.__array_finalize__(self)
            output._unit = output_units
            return output

        # Handle MaskedArray results
        if isinstance(arrays, np.ma.MaskedArray):
            mask = arrays.mask if arrays.mask is not np.ma.nomask else None
            output = arrays.data.view(type(self))
            output.__array_finalize__(self)
            output.mask = mask
            output._unit = output_units
            return output

        # Handle scalar results (no mask from MaskedArray, so no mask on output)
        if np.isscalar(arrays):
            output = type(self)(arrays, mask=None)
            output.__array_finalize__(self)
            output._mask = None  # Explicitly clear mask (not inherited from self)
            output._unit = output_units
            return output

        # Handle ndarray results (no mask from MaskedArray, so no mask on output)
        if isinstance(arrays, np.ndarray):
            output = arrays.view(type(self))
            output.__array_finalize__(self)
            output._mask = None  # Explicitly clear mask (not inherited from self)
            output._unit = output_units
            return output

        # For other results (sequences, etc.), delegate to parent
        return super()._convert_function_outputs(arrays, output_units, conversion)

    def __getitem__(self, key):
        item = super().__getitem__(key)
        if not isinstance(item, Tod):
            return item
        if item.mask is not None:
            item.mask = item.mask[key]
        return item

    @classmethod
    def empty(
        cls,
        shape,
        mask=None,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            empty(shape, dtype, order),
            mask=mask,
            header=header,
            unit=unit,
            derived_units=derived_units,
            dtype=dtype,
            copy=False,
            **keywords,
        )

    @classmethod
    def ones(
        cls,
        shape,
        mask=None,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.ones(shape, dtype, order),
            mask=mask,
            header=header,
            unit=unit,
            derived_units=derived_units,
            dtype=dtype,
            copy=False,
            **keywords,
        )

    @classmethod
    def zeros(
        cls,
        shape,
        mask=None,
        header=None,
        unit=None,
        derived_units=None,
        dtype=None,
        order=None,
        **keywords,
    ):
        if dtype is None:
            dtype = cls.default_dtype
        return cls(
            np.zeros(shape, dtype, order),
            mask=mask,
            header=header,
            unit=unit,
            derived_units=derived_units,
            dtype=dtype,
            copy=False,
            **keywords,
        )

    def imshow(
        self,
        mask=None,
        xlabel='Sample',
        ylabel='Detector number',
        aspect='auto',
        origin='upper',
        percentile=0,
        **keywords,
    ):
        if mask is None:
            mask = self.mask
        return super().imshow(
            mask=mask,
            xlabel=xlabel,
            ylabel=ylabel,
            aspect=aspect,
            origin=origin,
            percentile=percentile,
            **keywords,
        )

    def __str__(self):
        output = FitsArray.__str__(self)
        if self.ndim == 0:
            return output
        output += ' ('
        if self.ndim > 1:
            output += str(self.shape[-2]) + ' detector'
            if self.shape[-2] > 1:
                output += 's'
            output += ', '
        output += str(self.shape[-1]) + ' sample'
        if self.shape[-1] > 1:
            output += 's'
        return output + ')'

    def save(self, filename, comm=None):
        if comm is None:
            comm = MPI.COMM_SELF

        FitsArray.save(self, filename, comm=comm)
        if self.mask is not None:
            mask = self.mask.view('uint8')
            write_fits(filename, mask, None, True, 'Mask', comm)

    def ravel(self, order='C'):
        result = super().ravel(order)
        mask = None if self.mask is None else self.mask.ravel(order)
        result.mask = mask
        return result

    def reshape(self, *newdims, order='C'):
        result = super().reshape(*newdims, order=order)
        mask = None if self.mask is None else self.mask.reshape(*newdims, order=order)
        result.mask = mask
        return result

    def median(self, axis=None):
        #        result = Tod(median(self, mask=self.mask, axis=axis),
        result = Tod(
            np.median(self, axis=axis),
            header=self.header.copy(),
            unit=self.unit,
            derived_units=self.derived_units,
            dtype=self.dtype,
            copy=None,
        )
        result.mask = None
        #        # median might introduce NaN from the mask, let's remove them
        #        tmf.processing.filter_nonfinite_mask_inplace(result.ravel(),
        #            result.mask.view(np.int8).ravel())
        return result

    median.__doc__ = np.median.__doc__

    def sort(self, axis=-1, kind='quicksort', order=None):
        self.magnitude.sort(axis, kind, order)
        self.mask = None



from pathlib import Path

import numpy as np
import pytest
from astropy.io import fits as pyfits
from astropy.utils.exceptions import AstropyUserWarning
from astropy.wcs import WCS

from pyoperators.utils.testing import assert_eq, assert_same
from pysimulators.wcsutils import (
    WCSToWorldOperator,
    angle_lonlat,
    barycenter_lonlat,
    combine_fitsheader,
    create_fitsheader,
    create_fitsheader_for,
    fitsheader2shape,
    get_cdelt_pa,
    has_wcs,
    mean_degrees,
)

DATAPATH = Path(__file__).parent / 'data'


def test_mean_degrees():
    assert mean_degrees([1, 2]) == 1.5
    assert_same(mean_degrees([1, 359.1]), 0.05, atol=100)
    assert_same(mean_degrees([0.1, 359.1]), 359.6)


def test_angle_lonlat1():
    assert_same(angle_lonlat(30, 0, 40, 0), 10, atol=100)


@pytest.mark.parametrize(
    'c1, c2, angle',
    [
        ((30, 0), (40, 0), 10),
        ((39, 0), (92, 90), 90),
        ((39, 0), (37, -90), 90),
        ((37, -90), (39, 0), 90),
        ((100, 90), (134, -90), 180),
        ((24, 30), (24, 32), 2),
    ],
)
def test_angle_lonlat2(c1, c2, angle):
    assert_same(angle_lonlat(c1, c2), angle, rtol=1000)


def test_barycenter_lonlat():
    assert_eq(barycenter_lonlat([30, 40], [0, 0]), (35, 0))
    assert_eq(barycenter_lonlat([20, 20, 20], [-90, 0, 90]), (20, 0))
    assert_eq(barycenter_lonlat([20, 20, 20], [0, 45, 90]), (20, 45))


def test_get_cdelt_pa1():
    header = create_fitsheader((10, 10), cdelt=(-1.2, 3))
    cdelt, pa = get_cdelt_pa(header)
    assert_eq(cdelt, (-1.2, 3))
    assert_eq(pa, 0)


def test_get_cdelt_pa2():
    cdelt = (-1.5, 3)
    pa = -25.0
    header = create_fitsheader((10, 10), cdelt=cdelt, pa=pa)
    cdelt_, pa_ = get_cdelt_pa(header)
    assert_eq(cdelt, cdelt_)
    assert_eq(pa, pa_)


COMBINE_HEADERS = [
    create_fitsheader((1, 1), cdelt=3.0, crval=(0, 0)),
    create_fitsheader((3, 3), cdelt=1.0, crval=(1, 1)),
    create_fitsheader((5, 5), cdelt=1.0, crval=(3, 3)),
    create_fitsheader((2, 2), cdelt=1.0, crval=(5, 5)),
]
COMBINED_HEADER = combine_fitsheader(COMBINE_HEADERS)


@pytest.mark.parametrize('iheader, header', enumerate(COMBINE_HEADERS))
def test_combine_fitsheader(iheader, header):
    proj0 = WCS(COMBINED_HEADER)
    epsilon = 1.0e-10

    nx = header['NAXIS1']
    ny = header['NAXIS2']
    x = header['CRPIX1']
    y = header['CRPIX2']
    edges = (
        np.array(3 * (0.5, x, nx + 0.5)),
        np.array((0.5, 0.5, 0.5, y, y, y, ny + 0.5, ny + 0.5, ny + 0.5)),
    )
    a, d = WCS(header).wcs_pix2world(edges[0], edges[1], 1)
    x0, y0 = proj0.wcs_world2pix(a, d, 1)
    assert np.all(x0 > 0.5 - epsilon)
    assert np.all(y0 > 0.5 - epsilon)
    assert np.all(x0 < COMBINED_HEADER['NAXIS1'] + 0.5 + epsilon)
    assert np.all(y0 < COMBINED_HEADER['NAXIS2'] + 0.5 + epsilon)
    if iheader == 0:
        assert y0[0] <= 1.5 and x0[-1] >= COMBINED_HEADER['NAXIS1'] - 0.5
    if iheader == 3:
        assert x0[0] <= 1.5 and y0[-1] >= COMBINED_HEADER['NAXIS2'] - 0.5


def test_has_wcs():
    header = create_fitsheader([2, 4])
    assert not has_wcs(header)
    header = create_fitsheader([2, 4], cdelt=1)
    assert has_wcs(header)


@pytest.mark.parametrize('shape', [(), (1,), (1, 2), (0, 1), (1, 0), (2, 3)])
def test_create_fitsheader1(shape):
    array = np.ones(shape)

    h1 = create_fitsheader(
        array.shape[::-1], dtype=float if array.ndim > 0 else np.int32
    )
    h2 = create_fitsheader_for(array if array.ndim > 0 else None)
    assert h1 == h2
    assert h1['NAXIS'] == array.ndim
    for i, s in enumerate(array.shape[::-1]):
        assert h1['NAXIS' + str(i + 1)] == s


def test_create_fitsheader2():
    header = create_fitsheader_for(None)
    assert header['NAXIS'] == 0


def test_create_fitsheader3():
    header = create_fitsheader((10, 3), cdelt=1)
    assert_eq(header['RADESYS'], 'ICRS')
    assert 'EQUINOX' not in header


@pytest.mark.parametrize('naxes', [(), (1,), (1, 2), (2, 0, 3)])
def test_fitsheader2shape(naxes):
    header = create_fitsheader(naxes)
    assert_eq(fitsheader2shape(header), naxes[::-1])


@pytest.mark.parametrize('origin', [0, 1])
def test_wcsoperator(origin):
    with pytest.warns(AstropyUserWarning, match='File may have been truncated'):
        header = pyfits.open(DATAPATH / 'header_gnomonic.fits')[0].header

    crval = (header['CRVAL1'], header['CRVAL2'])
    crpix = np.array((header['CRPIX1'], header['CRPIX2'])) - 1 + origin
    toworld = WCSToWorldOperator(header, origin)
    assert_same(toworld(crpix), crval, atol=10)
    assert_same(toworld.I(crval), crpix, atol=100000)
